/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <ctype.h>
#include <string.h>
#include "muttlib.h"

static char *next_word (char *s)
{
  while (*s && *s != ' ') s++;
  while (*s && *s == ' ') s++;
  return s;
}

static int get_month (const char *s)
{
  switch(*s) {
  case 'j':
  case 'J':
    if (s[1] == 'a' || s[1] == 'A') {
      if (s[2] == 'n' || s[2] == 'N')
        return 0;
    }
    else if (s[1] == 'u') {
      if (s[2] == 'l' || s[2] == 'L')
        return 6;
      else if (s[2] == 'n' || s[2] == 'N')
        return 5;
    }
    break;
  case 'f':
  case 'F':
    if ((s[1] == 'e' || s[1] == 'E') && (s[2] == 'b' || s[2] == 'B'))
      return 1;
    break;
  case 'm':
  case 'M':
    if (s[1] == 'a' || s[1] == 'A') {
      switch(s[2]) {
        case 'y':
        case 'Y':
          return 4;
        case 'r':
        case 'R':
          return 2;
      }
    }
    break;
  case 'a':
  case 'A':
    switch(s[1]) {
    case 'p':
    case 'P':
      if (s[2] == 'r' || s[2] == 'R')
        return 3;
      break;
    case 'u':
    case 'U':
      if (s[2] == 'g' || s[2] == 'G')
        return 7;
      break;
    }
    break;
  case 's':
  case 'S':
    if ((s[1] == 'e' || s[1] == 'E') && (s[2] == 'p' || s[2] == 'P'))
      return 8;
    break;
  case 'o':
  case 'O':
    if ((s[1] == 'c' || s[1] == 'C') && (s[2] == 't' || s[2] == 'T'))
      return 9;
    break;
  case 'n':
  case 'N':
    if ((s[1] == 'o' || s[1] == 'O') && (s[2] == 'v' || s[2] == 'V'))
      return 10;
    break;
  case 'd':
  case 'D':
    if ((s[1] == 'e' || s[1] == 'E') && (s[2] == 'c' || s[2] == 'C'))
      return 11;
    break;
  }
  return(-1);
}

static int is_day_name (char *s)
{
  if (!whitespace(*(s+3))) return 0;
  if (strncasecmp("mon", s, 3) == 0)
    return 1;
  if (strncasecmp("tue", s, 3) == 0)
    return 1;
  if (strncasecmp("wed", s, 3) == 0)
    return 1;
  if (strncasecmp("thu", s, 3) == 0)
    return 1;
  if (strncasecmp("fri", s, 3) == 0)
    return 1;
  if (strncasecmp("sat", s, 3) == 0)
    return 1;
  if (strncasecmp("sun", s, 3) == 0)
    return 1;
  return 0;
}

/* A valid message separator looks like:
 *
 * From [ <return-path> ] <weekday> <month> <day> <time> [ <timezone> ] <year>
 */

time_t is_from (char *s)
{
  struct tm tm;
  int yr;
  char *p;

  if (strncmp("From ", s, 5) != 0) return 0;

  s = next_word(s); /* skip over the From part. */
  if (!*s) return 0;

  dprint(3, (debugfile, "\nis_from(): parsing: %s", s));

  if (!is_day_name(s)) {

    /* looks like we got the return-path, now fetch the weekday */
    s = next_word(s);
    if (!*s) return 0;

    if (!is_day_name(s)) {
      dprint(1, (debugfile, "is_from():  expected weekday, got: %s\n", s));
      return 0;
    }
  }

  s = next_word(s);

  /* do a quick check to make sure that this isn't really the day of the week.
   * this could happen when receiving mail from a local user whose login name
   * is the same as a three-letter abbreviation of the day of the week.
   */
  if (is_day_name(s)) {
    s = next_word(s);
    /* now we should be on the month. */
  }

  if (!*s) return 0;
  if ((tm.tm_mon = get_month(s)) < 0) return 0;

  /* day */
  s = next_word(s);
  if (!*s) return 0;

  p = s;
  s = next_word(s);
  if (!*s) return 0;

  if (sscanf(p, "%d", &tm.tm_mday) != 1) return 0;

  /* time */
  p = s;
  s = next_word(s);
  if (!*s) return 0;

  /* Accept either HH:MM or HH:MM:SS */
  if (sscanf(p, "%d:%d:%d", &tm.tm_hour, &tm.tm_min, &tm.tm_sec) == 3);
  else if (sscanf(p, "%d:%d", &tm.tm_hour, &tm.tm_min) == 2) {
    tm.tm_sec = 0;
  }
  else
    return 0;

  /* timezone? */
  p = s;
  s = next_word(s);

  if (isalpha(*p) || *p == '+' || *p == '-') {
    if (!*s) return 0;
    p = s;
    s = next_word(s);
  }

  /* year */
  if (sscanf(p, "%d", &yr) != 1) return 0;
  tm.tm_year = yr > 1900 ? yr - 1900 : yr;

  dprint(3,(debugfile, "is_from(): month=%d, day=%d, hr=%d, min=%d, sec=%d, yr=%d.\n",
            tm.tm_mon, tm.tm_mday, tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_year));

  tm.tm_isdst = 0;
  tm.tm_yday = 0;
  tm.tm_wday = 0;

  return(mktime(&tm));
}
