/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <pwd.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/utsname.h>
#include <stdio.h>
#include <time.h>
#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"

#define DT_BOOL 1
#define DT_STR 2
#define DT_NUM 3

#define NIL 0

struct rc_option {
  char *option;
  short type;
  void *data;
  long bit; /* DT_BOOL only */
} rc_vars[] = {
  { "allow_8bit",   DT_BOOL, 0, OPTALLOW8BIT },
  { "arrow_cursor", DT_BOOL, 0, OPTARROWCURSOR },
  { "askcc",        DT_BOOL, 0, OPTASKCC },
  { "askcopy",      DT_BOOL, 0, OPTASKCOPY },
  { "askdelete",    DT_BOOL, 0, OPTASKDEL },
  { "askmove",      DT_BOOL, 0, OPTASKMOVE },
  { "askpostpone",  DT_BOOL, 0, OPTASKPOSTPONE },
  { "askreplyto",   DT_BOOL, 0, OPTASKREPLYTO },
  { "attribution",  DT_STR,  Attribution , 0 },
  { "charset",      DT_STR,  Charset, 0 },
  { "domain",       DT_STR,  Domain, 0 },
  { "edit_hdrs",    DT_BOOL, 0, OPTEDITHDRS },
  { "editor",       DT_STR,  Editor },
  { "folder",       DT_STR,  Maildir },
  { "force_name",   DT_BOOL, 0, OPTFORCENAME },
  { "floating_prompt", DT_BOOL, 0, OPTFLOATINGPROMPT },
  { "hdr_format",   DT_STR, Hdr_format, 0 },
  { "header",       DT_BOOL, 0, OPTHEADER },
  { "hold",         DT_BOOL, 0, OPTHOLD },
  { "honor_length", DT_BOOL, 0, OPTHONORLENGTH },
  { "in_reply_to",  DT_STR, In_reply_to, 0 },
  { "indentprefix", DT_STR, Prefix, 0 },
  { "keep_empty",   DT_BOOL, 0, OPTKEEPEMPTY },
  { "keepsave",     DT_BOOL, 0, OPTKEEPSAVE },
  { "localsignature", DT_STR, Local_sig, 0 },
  { "mbox",         DT_STR, Inbox, 0 },
  { "metoo",        DT_BOOL, 0, OPTMETOO },
  { "mime_forward", DT_BOOL, 0, OPTMIMEFORWARD },
  { "msginc",       DT_NUM, &Msg_inc, 0 },
  { "pager",        DT_STR, Pager, 0 },
#ifdef _PGPPATH
  { "pgp",          DT_STR, Pgp, 0 },
#endif /* _PGPPATH */
  { "pointnew",     DT_BOOL, 0, OPTPOINTNEW },
  { "post_indent_str", DT_STR, PostIndentString, 0 },
  { "postponed",    DT_STR, Postponed, 0 },
  { "print",        DT_STR, Print, 0 },
  { "prompt_after", DT_BOOL, 0, OPTPROMPTAFTER },
  { "realname",     DT_STR, Realname, 0 },
  { "record",       DT_STR, Outbox, 0 },
  { "remotesignature", DT_STR, Remote_sig, 0 },
  { "reverse_alias", DT_BOOL, 0, OPTREVALIAS },
  { "sendmail",     DT_STR, Sendmail, 0 },
  { "show_host",    DT_BOOL, 0, OPTSHOWHOST },
  { "sigdashes",    DT_BOOL, 0, OPTSIGDASHES },
  { "signature",    DT_STR, Signature, 0 },
  { "spoolfile",    DT_STR, Spoolfile, 0 },
  { "xterm_mouse",  DT_BOOL, 0, OPTXTERMMOUSE },
  { 0, 0, 0 }
};

char Init_error[SHORT_STRING];

static int option_index (char *s) {
  int i;

  for (i=0; rc_vars[i].option; i++) {
    if (strcmp(s, rc_vars[i].option) == 0)
      return(i);
  }
  return(-1);
}

static void unignore_header (char *s)
{
  LIST *p = Ignore, *last=NULL;
  
  if (strcmp("*", s) == 0) {
    while (Ignore) {
      p=Ignore;
      Ignore=Ignore->next;
      safe_free((void **)&p->data);
      safe_free((void **)&p);
    }
    return;
  }
  while (p) {
    if (strcasecmp(s, p->data) == 0) {
      safe_free((void **)&p->data);
      if (last)
	last->next = p->next;
      else
	Ignore = p->next;
      safe_free((void **)&p);
    }
    last = p;
    p = p->next;
  }
}

static void parse_unignore (char *s)
{
  while ((s = strtok(s, " \t")) != NULL) {
    unignore_header(s);
    s=NULL;
  }
}

static void parse_alternates (char *s) {
  ADDRESS *end = Alternates;
  ADDRESS *newaddr = NIL;
  char *p;

  if (end)
    while (end->next) end=end->next;

  p = s;
  while ((p = strpbrk(p, " \t")) != NULL)
    *p = ',';
    
  rfc822_parse_adrlist(&newaddr, s, Hostname);
  if (end)
    end->next = newaddr;
  else
    Alternates = newaddr;
}

static void add_userheader(char *s) {
  USERHEADER *new = (USERHEADER *)safe_malloc(sizeof(USERHEADER));
  char *p;
  char buffer[STRING];

  p = strchr(s, ':');
  *p = 0;
  p++;
  while (*p && isspace(*p)) p++;
  new->key = safe_strdup(s);
  new->val = safe_strdup(p);
  rfc1522_encode_string (buffer, (unsigned char *)new->val, sizeof(buffer));
  safe_free((void **)&new->val);
  new->val = safe_strdup(buffer);
  new->next = NIL;

  if (User_defined_headers) {
    USERHEADER *tmp = User_defined_headers;

    while (tmp->next) tmp = tmp->next;
    tmp->next = new;
  }
  else
    User_defined_headers = new;
}

static void add_alias (char *s) {
  ALIAS *new = (ALIAS *)safe_malloc(sizeof(ALIAS));
  char *p;

  p = strpbrk(s, " \t");
  *p = 0;
  p++;
  while (*p && isspace(*p)) p++;

  new->name = safe_strdup(s);
  new->addr = 0;
  rfc822_parse_adrlist(&new->addr, p, "@");
  new->next = NIL;
  if (Aliases == NIL)
      Aliases = new;
  else {
    ALIAS *tmp = Aliases;

    while (tmp->next) tmp = tmp->next;
    tmp->next = new;
  }
}

void mutt_add_to_list (LIST **list, char *s) {
  LIST *t;

  while ((s = strtok(s, " \t,"))) {
    t = (LIST *)safe_malloc(sizeof(LIST));
    t->data = safe_strdup(s);
    t->next = *list;
    *list = t;
    s = NULL;
  }
}

/*
 * reads the specified initialization file.  returns -1 if errors were found
 * so that we can pause to let the user know...
 */
static int source_rc (char *rcfile) {
  FILE *f;
  int line = 0, unset = 0, rc = 0;
  char *pk, keyword[SHORT_STRING],  buffer[LONG_STRING], *p;

  mutt_expand_path(rcfile);

#ifdef USE_SETGID
  if (access(rcfile, R_OK) == -1) {
    printf("%s: permission denied.\n", rcfile);
    return(-1);
  }
#endif /* USE_SETGID */

  if ((f = fopen(rcfile, "r")) == NULL) {
    fprintf(stderr, "%s: no such file or directory.", rcfile);
    return(-1);
  }
  while (fgets(buffer, LONG_STRING, f) != NULL) {
    line++;
    
    /* Strip those comments! */
    if ((p = strchr(buffer, '#'))) *p = 0;
    
    /* Remove any trailing whitespace */
    p = buffer + strlen(buffer) - 1;
    while (isspace(*p) && (p >= buffer)) {
      *p = 0;
      if (p == buffer) break;
      else p--;
    }
    
    p = buffer;
    while (*p && isspace(*p)) p++;
    
    /* Is there anything left on this line? */
    if (!*p) continue;
    
    if (strncmp(p, "set", 3) == 0 ||
        (strncmp(p, "unset", 5) == 0 && (unset=1))) {
      int idx;

      p = strpbrk(p, " \t");
      while (*p && isspace(*p)) p++;
      
      /* Get the keyword */
      pk = keyword;
      while (*p && !isspace(*p) && *p != '=') *pk++ = *p++;
      *pk = 0;

      idx = option_index(keyword);
      if (idx == -1) {
        printf("Error in %s, line %d: %s: unknown variable.\n", rcfile,
               line, keyword);
        rc = -1;
        continue;
      }
      if (rc_vars[idx].type == DT_BOOL) {
        if (unset)
          Options &= ~(rc_vars[idx].bit);
        else
          Options |= rc_vars[idx].bit;
      }
      else if (rc_vars[idx].type == DT_STR) {
        char *ptr = rc_vars[idx].data;

        if (*p != '=') {
          printf("Error in %s, line %d: expected '=' but got %s\n", rcfile,
                 line, p);
          rc = -1;
          continue;
        }
        p++;
        if (*p == '"') {
          p++;
          while (*p && *p != '"') *ptr++ = *p++;
          *ptr = 0;
        }
        else {
          while (*p && !isspace(*p)) *ptr++ = *p++;
          *ptr = 0;
        }
      }
      else if (rc_vars[idx].type == DT_NUM) {
        short *ptr = rc_vars[idx].data;

        if (*p != '=') {
          printf("Error in %s, line %d: expected '=' but got %s\n", rcfile,
                 line, p);
          rc = -1;
          continue;
        }
        p++;
        while (*p && isspace(*p)) p++;
        *ptr = atoi(p);
      }
      unset = 0;
    }
    else if (strncmp(p, "alternates", 10) == 0) {
      p = skip_whitespace(p+10);
      parse_alternates(p);
    }
    else if (strncmp(p, "ignore", 6) == 0) {
      p = skip_whitespace(p+6);
      mutt_add_to_list(&Ignore, p);
    }
    else if (strncmp(p, "unignore", 8) == 0) {
      p = skip_whitespace(p+8);
      parse_unignore(p);
    }
    else if (strncmp(p, "my_hdr", 6) == 0) {
      p = skip_whitespace(p+6);
      add_userheader(p);
    }
    else if (strncmp(p, "alias", 5) == 0) {
      p = skip_whitespace(p+5);
      add_alias(p);
    }
    else if (strncmp(p, "lists", 5) == 0) {
      p = skip_whitespace(p+5);
      mutt_add_to_list(&Mailing_lists, p);
    }
    else if (strncmp(p, "source", 6) == 0) {
      p = skip_whitespace(p+6);
      if (source_rc(p) == -1) rc = -1;
    }
    else if (strncmp(p, "sort", 4) == 0) {
      Sort = 0;
      p = skip_whitespace(p+4);
      if (strncasecmp(p, "reverse-", 8) == 0) {
        Sort = SORTREVERSE;
        p += 8;
      }
      if (strncasecmp(p, "date-sent", 9) == 0) Sort |= SORTDATE;
      else if (strncasecmp(p, "date-received", 13) == 0) Sort |= SORTORDER;
      else if (strncasecmp(p, "from", 4) == 0) Sort |= SORTFROM;
      else if (strncasecmp(p, "subject", 7) == 0) Sort |= SORTSUBJECT;
      else if (strncasecmp(p, "size", 4) == 0) Sort |= SORTSIZE;
      else {
        printf("Error in %s, line %d: unknown sorting method: %s\n",
               rcfile, line, p);
        Sort = SORTDATE;
        rc = -1;
      }
    }
    else if (strncmp(p, "bind", 4) == 0) {
      p = skip_whitespace(p+4);
      if (mutt_parse_bind(p) == -1) {
        printf("Error in %s, line %d: %s\n", rcfile, line, Init_error);
        rc = -1;
      }
    }
    else {
      char *c = strpbrk(p, " \t");

      if (c) *c = 0;
      printf("Error in %s, line %d: %s: unknown command.\n", rcfile, line, p);
      rc = -1;
    }
  }
  fclose(f);
  return(rc);
}

int mutt_init (int skip_sys_rc) {
  struct passwd *pw;
  char buffer[STRING];
  char *p;
  int code = 0;
#ifndef HOSTNAME
  struct utsname utsname;
#endif
#ifdef DEBUG
  time_t t;
#endif

  /* Get some information about the user */
  pw = getpwuid(getuid());
  strfcpy(Username, pw->pw_name, sizeof(Username));
  strfcpy(Homedir, pw->pw_dir, sizeof(Homedir));
  strfcpy(Realname, pw->pw_gecos, sizeof(Realname));
  if ((p = strchr(Realname, ',')) != 0) *p = 0;

  /* Start up debugging mode if requested */
#ifdef DEBUG
  if (debuglevel > 0) {
    sprintf(buffer, "%s/.muttdebug", Homedir);
    debugfile = fopen(buffer, "w");
    t = time(0);
    fprintf(debugfile, "%s started at %s.\nDebugging at level %d.\n\n",
            Mutt_version, asctime(localtime(&t)), debuglevel);
  }
#endif

  /* And about the host... */
#ifdef HOSTNAME
  strcpy(Hostname, HOSTNAME);
#else
  uname(&utsname);
  strcpy(Hostname, utsname.nodename);
  /* some systems report the FQDN instead of just the hostname */
  if ((p = strchr(Hostname, '.')))
    *p = 0;
#endif /* HOSTNAME */

#ifdef USE_DOMAIN
  strcpy(Domain, DOMAIN);
#endif

  /* Set some defaults */

  strcpy(Tempdir, "/tmp");
  strcpy(Pager, "builtin");  
  strcpy(Maildir, "~/Mail");
  strcpy(Inbox, "~/mbox");
  strcpy(Postponed, "~/postponed");
  strcpy(Outbox, "");

  if ((p = getenv("MAIL")))
    strcpy(Spoolfile, p);
  else {
#ifdef HOMESPOOL
    sprintf(Spoolfile, "%s/%s", Homedir, MAILPATH);
#else
    sprintf(Spoolfile, "%s/%s", MAILPATH, Username);
#endif
  }

  strcpy(Sendmail, SENDMAIL);
  strcpy(Print, "lpr");
/* default charset is now euc-kr - pms@romance.kaist.ac.kr */
  strcpy(Charset, "euc-kr");
  if ((p = getenv("VISUAL")) != NULL)
    strcpy(Editor, p);
  else if ((p = getenv("EDITOR")) != NULL)
    strcpy(Editor, p);
  else
    strcpy(Editor, "vi");
  strcpy(Attribution, "%n writes:");
  strcpy(Prefix, "> ");
  PostIndentString[0] = 0;
  strcpy(Signature, "~/.signature");
  strcpy(Local_sig, "");
  strcpy(Remote_sig, "");
  strcpy(In_reply_to, "%i; from %n on %d");
  strcpy(Hdr_format, "%M %2N %-15.15L (%6c) %s");
  Metamail[0] = 0;
#ifdef _PGPPATH
  strcpy(Pgp, _PGPPATH);
#endif /* _PGPPATH */

  /*
   * Process the global rc file if it exists and the user hasn't explicity
   * requested not to via "-n".
   */
  if (!skip_sys_rc && access(SYSMUTTRC, F_OK) != -1) source_rc(SYSMUTTRC);

  /*
   * Read the user's initialization file.
   */
  sprintf(buffer, "%s/.muttrc", Homedir);
  if (access(buffer, F_OK) != -1)
    code = source_rc(buffer);

  /*
   * Special case which needs handling up front since expand_path() is not
   * recursive.
   */
  if (Maildir[0] == '~') {
    sprintf(buffer, "%s/%s", Homedir, Maildir+2);
    strcpy(Maildir, buffer);
  }

  /* expand these now so we don't have to keep doing it over and over */
  mutt_expand_path(Inbox);
  mutt_expand_path(Outbox);
  mutt_expand_path(Postponed);
  mutt_expand_path(Spoolfile);
  mutt_expand_path(Local_sig);
  mutt_expand_path(Remote_sig);
  mutt_expand_path(Signature);

  /*
   * The user may have changed the domain name in an init file, so wait
   * until now to set our FQDN for outgoing mail.
   */
#ifdef USE_DOMAIN
# ifdef HIDDEN_HOST
  strcpy(Fqdn, Domain);
# else
  sprintf(Fqdn, "%s.%s", Hostname, Domain);
# endif /* HIDDEN_HOST */
#else
  strcpy(Fqdn, Hostname);
#endif

  return code;
}
