/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/wait.h>
#include "muttlib.h"
#include "rfc1522.h"

static char *compatible_charsets[] = {
  "US-ASCII",
/* euc-kr and iso-2022-kr are added - pms@romance.kaist.ac.kr */
  "EUC-KR",
  "ISO-2022-KR",
  "ISO-8859-1",
  "ISO-8859-2",
  "ISO-8859-3",
  "ISO-8859-4",
  "ISO-8859-5",
  "ISO-8859-6",
  "ISO-8859-7",
  "ISO-8859-8",
  "ISO-8859-9",
  "KOI8-R",
  (char *)0,
};

int compat_charset (const char *s)
{
  int i;

  for (i=0; compatible_charsets[i]; i++)
    if (strcasecmp(s, compatible_charsets[i]) == 0)
      return(1);
  return(0);
}

METAENV *mutt_new_metaenv (void)
{
  METAENV *p = (METAENV *)safe_malloc(sizeof(METAENV));

  p->env = NULL;
  p->userhdrs[0] = NULL;
  return(p);
}

BODY *mutt_new_body (void)
{
  BODY *p = (BODY *)safe_malloc(sizeof(BODY));
    
  p->type = 0;
  p->encoding = 0;
  p->subtype = 0;
  p->parameter = 0;
  p->description = 0;
  p->filename = 0;
  p->content = 0;
  p->md5 = 0;
  p->next = 0;
  p->disposition = DISPATTACH;
  p->parts = 0;
  p->use_disp = 1;
  p->unlink = 0; /* important! */
  p->hdr_offset = p->offset = p->length = 0;
  return(p);
}

void mutt_free_body (BODY **p)
{
  BODY *a = *p, *b;

  while (a) {
    b = a;
    a = a->next; 
    if (b->parameter) mutt_free_parameter(&b->parameter);
    if (b->unlink) unlink(b->filename);
    safe_free((void **)&b->filename);
    safe_free((void **)&b->content);
    safe_free((void **)&b->subtype);
    safe_free((void **)&b->description);
    safe_free((void **)&b);
  }
  *p = 0;
}

PARAMETER *mutt_new_parameter (void)
{
  PARAMETER *p = (PARAMETER *)safe_malloc(sizeof(PARAMETER));
  p->attribute = 0;
  p->value = 0;
  p->next = 0;
  return(p);
}

void mutt_free_parameter (PARAMETER **p)
{
  PARAMETER *t = *p;
  PARAMETER *o;

  while (t) {
    safe_free((void **)&t->attribute);
    safe_free((void **)&t->value);
    o = t;
    t = t->next;
    safe_free((void **)&o);
  }
  *p = 0;
}

STATE *mutt_new_state (void)
{
  STATE *p;
  
  p = (STATE *)safe_malloc(sizeof(STATE));
  p->fpin = p->fpout = 0;
  p->displaying = 0;
  p->prefix = 0;
  p->offset = 0;
  p->array = 0;
  return p;
}

void mutt_free_state (STATE **p)
{
  safe_free((void **)p);
}

HEADER *mutt_new_header (void)
{
  HEADER *p = (HEADER *)safe_malloc(sizeof(HEADER));

  p->from = p->to = p->cc = p->sender = p->reply_to = 0;
  p->subject = p->message_id = 0;
  p->content = 0;
  p->references = 0;

  p->offset = p->length = 0;
  p->lines = p->msgno = p->virtual = 0;

  /* Clear all the flags associated with this message */
  p->read = p->selected = p->replied = p->old = p->deleted = p->flagged = 0;
  p->changed = p->marked = p->remail = p->mime = p->metamail = 0;
#ifdef _PGPPATH
  p->pgp = 0;
#endif /* _PGPPATH */

  p->year = 0;
  p->month = 0;
  p->day = 0;
  p->hours = 0;
  p->minutes = 0;
  p->seconds = 0;
  p->zhours = 0;
  p->zminutes = 0;
  p->zoccident = 0;

  p->received = 0;

  return(p);
}

void mutt_free_header (HEADER **h)
{
  mutt_free_address(&(*h)->from);
  mutt_free_address(&(*h)->to);
  mutt_free_address(&(*h)->cc);
  mutt_free_address(&(*h)->reply_to);
  mutt_free_address(&(*h)->sender);
  mutt_free_body(&(*h)->content);
  safe_free((void **)&(*h)->subject);
  safe_free((void **)&(*h)->message_id);
  safe_free((void **)&(*h)->references);
  *h = 0;
}

void remove_leading_space (char *s)
{
  char *p = s;

  while (*p && isspace(*p)) p++;
  if (p != s && *p) {
    while (*p) *s++ = *p++;
  }
}

/* returns true if the given header field is named in the user's ignore list */
int mutt_matches_ignore (const char *s)
{
  LIST *t = Ignore;

  while (t) {
    if (strncasecmp(s, t->data, strlen(t->data)) == 0)
      return 1;
    t = t->next;
  }
  return 0;
}

/*
 *  This routine copies a RFC822 header from the current offset in "f" to the
 * state "s".  It weeds out any headers which the user does not care to see,
 * and also takes care of decoding any RFC1522 encoded words encountered.
 */
void copy_header (STATE *s, int ignoring)
{
  char buffer[LONG_STRING];
  int ignore = 0, class = 0;

  /* slurp the message separator.  it _never_ gets printed. */
  if (fgets(buffer, sizeof(buffer), s->fpin) == NULL) return;

  while (fgets(buffer, sizeof(buffer), s->fpin) != NULL) {
    if (buffer[0] == '\n' || (buffer[0] == '\r' && buffer[1] == '\n')) break;
    if (whitespace(buffer[0])) {
      if (!ignore) {
        rfc1522_decode(buffer, sizeof(buffer), class);
        if (s->prefix) state_puts(s->prefix, s);
        state_puts(buffer, s);
      }
    }
    else if (ignoring && mutt_matches_ignore(buffer))
      ignore = 1;
    else {
      ignore = 0;
      class = rfc1522_classify(buffer);
      rfc1522_decode(buffer, sizeof(buffer), class);
      if (s->prefix) state_puts(s->prefix, s);
      state_puts(buffer, s);
    }
  }
  if (s->prefix) state_puts(s->prefix, s);
  state_puts(buffer, s);
}

void mutt_expand_path (char *s)
{
  char buf[_POSIX_PATH_MAX];

  if (*s == '~')
    sprintf(buf, "%s%s", Homedir, s+1);
  else if (*s == '=' || *s == '+')
    sprintf(buf, "%s/%s", Maildir, s+1);
  else if (strcmp(s, ">") == 0)
    strcpy(buf, Inbox);
  else if (strcmp(s, "<") == 0)
    strcpy(buf, Outbox);
  else if (strcmp(s, "!") == 0)
    strcpy(buf, Spoolfile);
  else
    return;
  strcpy(s, buf); /* replace the string with the expanded version. */
}

/* returns TRUE if the given address belongs to the user. */
int mutt_addr_is_user (ADDRESS *addr)
{
  ADDRESS *a = Alternates;

  if (!addr || !addr->mailbox) return 0;
  if (strcasecmp(addr->mailbox, Username) == 0 &&
      (strcasecmp(addr->host, Hostname) == 0 ||
       strcasecmp(addr->host, Fqdn) == 0))
    return 1;
  while (a) {
    if (strcasecmp(addr->mailbox, a->mailbox) == 0 && strcasecmp(addr->host, a->host) == 0)
      return 1;
    a = a->next;
  }
  return (0);
}

void *safe_malloc (unsigned int siz)
{
  void *p;

  if (siz == 0) return 0;
  if ((p = (void *)malloc(siz)) == 0) {
    mutt_error("Out of memory!");
    sleep(1);
    mutt_exit(1);
  }
  return(p);
}

void safe_realloc (void **p, unsigned int siz)
{
  void *r;

  if (siz == 0) {
    safe_free((void **)&*p);
    return;
  }
  if ((r = (void *)realloc(*p, siz)) == 0) {
    mutt_error("Out of memory!");
    sleep(1);
    mutt_exit(1);
  }
  *p = r;
}

void safe_free (void **p)
{
  if (!*p) return;
  free(*p);
  *p = 0;
}

char *safe_strdup (const char *s)
{
  char *p;

  if (!s) return 0;
  p = (char *)safe_malloc(strlen(s)+1);
  strcpy(p, s);
  return(p);
}

char *mutt_new_tempfile (void)
{
  char buffer[256];

  sprintf(buffer, "%s/mutt%d", Tempdir, (int)(LRAND()) % 1000000);
  return(safe_strdup(buffer));
}

char *skip_whitespace (char *p)
{
  while (*p && (*p == ' ' || *p == '\t')) p++;
  return(p);
}

void mutt_copy_bytes (FILE *in, FILE *out, long size)
{
  char buffer[LONG_STRING];
  long chunk;

  while (size > 0) {
    if (size > sizeof(buffer))
      chunk = sizeof(buffer);
    else
      chunk = size;
    if ((chunk = fread(buffer, 1, chunk, in)) < 1) break;
    fwrite(buffer, 1, chunk, out);
    size -= chunk;
  }
}

ADDRESS *mutt_new_address (void)
{
  ADDRESS *p = (ADDRESS *)safe_malloc(sizeof(ADDRESS));
  p->personal = 0;
  p->mailbox = 0;
  p->adl = 0;
  p->host = 0;
  p->error = 0;
  p->next = 0;
  return(p);
}

void mutt_free_address (ADDRESS **p)
{
  ADDRESS *t = *p;
  ADDRESS *o;

  while (t) {
    safe_free((void **)&t->personal);
    safe_free((void **)&t->mailbox);
    safe_free((void **)&t->adl);
    safe_free((void **)&t->host);
    safe_free((void **)&t->error);
    o = t;
    t = t->next;
    safe_free((void **)&o);
  }
  *p = 0;
}

char *mutt_get_parameter (const char *s, PARAMETER *p)
{
  for (; p; p=p->next)
    if (strcasecmp(s, p->attribute) == 0)
      return(p->value);
  return(0);
}

#ifdef METAMAIL
int mutt_needs_metamail (BODY *m)
{
  switch (m->type) {
  case TYPETEXT:
  case TYPEMULTIPART:
  case TYPEMESSAGE:
    return 0;
  default:
    return 1;
  }
}
#endif /* METAMAIL */

int mutt_is_text_type (int t, char *s)
{
  if (t == TYPETEXT)
    return 1;
  if (t == TYPEMESSAGE) {
    if (strcasecmp("delivery-status", s) == 0)
      return 1;
  }
  return 0;
}

/*
**  - modified by pms@romance.kaist.ac.kr
*/
void mutt_default_savename (char *s, HEADER *cur)
{
  ADDRESS *a = cur->to;

  *s++ = '=';
  
  /* First check to see whether or not this message is addressed to one of
   * the user's known mailing lists! */
  while (a) {
    if (mutt_is_mail_list(a)) {
      strcpy(s, a->mailbox);
      return;
    }
    a = a->next;
  }
  a = cur->cc;
  while (a) {
    if (mutt_is_mail_list(a)) {
      strcpy(s, a->mailbox);
      return;
    }
    a = a->next;
  }
  if (cur->reply_to)
    strcpy(s, cur->reply_to->mailbox);
  else if(cur->from)
    strcpy(s, cur->from->mailbox);
  else
    strcpy(s, cur->from->mailbox);
}

ADDRESS *mutt_default_from (void)
{
  ADDRESS *adr = mutt_new_address();

  if (Realname[0]) adr->personal = safe_strdup(Realname);
  adr->mailbox = safe_strdup(Username);
  adr->host = safe_strdup(Fqdn);
  return adr;
}

ENVELOPE *mutt_new_envelope (void)
{
  ENVELOPE *p = (ENVELOPE *)safe_malloc(sizeof(ENVELOPE));

  p->to = p->cc = p->bcc = p->sender = p->from = 0;
  p->subject = p->in_reply_to = 0;
  p->references = 0;
  return(p);
}

void mutt_free_envelope (ENVELOPE **p)
{
  if (!*p) return;
  mutt_free_address(&(*p)->to);
  mutt_free_address(&(*p)->cc);
  mutt_free_address(&(*p)->bcc);
  mutt_free_address(&(*p)->sender);
  mutt_free_address(&(*p)->from);
  safe_free((void **)&(*p)->subject);
  safe_free((void **)&(*p)->in_reply_to);
  safe_free((void **)&(*p)->references);
  safe_free((void **)p);
}

char *mutt_make_date (char *s)
{
  time_t t = time(0);
  struct tm *l = gmtime(&t);
  int yday = l->tm_yday;
  int tz = l->tm_hour * 60 + l->tm_min;

  l = localtime(&t);
  tz = l->tm_hour * 60 + l->tm_min - tz;
  yday = l->tm_yday - yday;

  if (yday != 0)
    tz += yday * 24 * 60; /* GMT is next or previous day! */

  sprintf(s, "Date: %s, %d %s %d %02d:%02d:%02d %+03d%02d\n",
    Weekdays[l->tm_wday], l->tm_mday, Months[l->tm_mon], l->tm_year+1900,
    l->tm_hour, l->tm_min, l->tm_sec, tz/60, abs(tz) % 60);
  return s;
}

void mutt_pipe_message (HEADER *h, const char *cmd)
{
  STATE *s;
  int status;

  s = mutt_new_state();
  mutt_create_filter(cmd, &s->fpout, NULL);
  s->fpin = Context->folder->fp;
  s->displaying = 0;
  fseek(s->fpin, h->offset, 0);
  copy_header(s, 1);
  mutt_body_handler(h->content, s);
  fclose(s->fpout);
  mutt_free_state(&s);
  wait(&status);
}

/* this does NOT copy the newline that separates the header from the body. */
void mutt_copy_header (HEADER *h, FILE *f, int xmit)
{
  int ignore=0, class=0;
  char buffer[LONG_STRING];

  fseek(MAILFP, h->offset, 0);
  while (fgets(buffer, sizeof(buffer)-1, MAILFP) != NULL) {
    if (strncmp("From ", buffer, 5) == 0)
      continue;
    else if (buffer[0] == '\n')
      return;
    else if (xmit) {
      if (strncmp("Content-Length", buffer, 14) == 0 ||
	  strncmp("Status", buffer, 6) == 0 ||
	  strncmp("X-Status", buffer, 8) == 0)
	continue;
      fputs(buffer, f);
    }
    else {
      if (Context->weed) {
	if (mutt_matches_ignore(buffer))
	  ignore = 1;
	else if (buffer[0] != ' ' && buffer[0] != '\t')
	  ignore = 0;
      }
      if (!ignore) {
	if (buffer[0] != ' ' && buffer[0] != '\t')
	  class = rfc1522_classify(buffer);
	rfc1522_decode(buffer, sizeof(buffer)-1, class);
	fputs(buffer, f);
      }
    }
  }
}

void mutt_tabs_to_spaces (char *s)
{
  while (*s) {
    if (*s == '\t')
      *s=' ';
    s++;
  }
}
