/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#define MAIN_C 1

#include <string.h>
#include <stdlib.h>
#include <locale.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <stdarg.h>
#include "mutt.h"
#include "rfc822.h"
#include "mutt_curses.h"
#include "version.h"
#include "copyright.h"

#ifdef USE_SETGID
#include <grp.h>
#endif /* USE_SETGID */

#include "keymap_global.c"

#ifdef NCURSES_MOUSE_VERSION
mmask_t Mousemask;
#endif

void mutt_error (const char * fmt, ...)
{
  va_list ap;

  va_start(ap, fmt);
  vsprintf(Errorbuf, fmt, ap);
  va_end(ap);
  
  move(LINES-1, 0);
  clrtoeol();
  CenterLine(Errorbuf, LINES-1);
  refresh();
}

void mutt_exit (int code)
{
  endwin();
  exit(code);
}

int mutt_system (const char *s)
{
  RETSIGTYPE (*h)(int);
  RETSIGTYPE (*i)(int);
  RETSIGTYPE (*t)(int);
  RETSIGTYPE (*c)(int);
  RETSIGTYPE (*term)(int);
  int rc;

  endwin();

  /* Reset the following signals so the subprocess can catch them. */
  h = signal(SIGHUP, SIG_DFL);
  i = signal(SIGINT, SIG_DFL);
  t = signal(SIGTSTP, SIG_DFL);
  c = signal(SIGCONT, SIG_DFL);
  term = signal(SIGTERM, SIG_DFL);

#ifdef USE_SETGID
  setegid(User_gid);
#endif /* USE_SETGID */

  rc = system(s);

#ifdef USE_SETGID
  setegid(Mail_gid);
#endif /* USE_SETGID */

  /* Restore our signal catchers */
  signal(SIGHUP, h);
  signal(SIGINT, i);
  signal(SIGTSTP, t);
  signal(SIGCONT, c);
  signal(SIGTERM, term);

  return(rc);
}

static void mutt_help (void)
{
  puts("\nUsage:\n\
   hanmutt [ -n ] [ -f <folder> ]\n\
   hanmutt [ -n ] [ -i <file> ] [ -s <subject> ] [ -c <addr> ] <addr> [ ... ]\n");
  exit(0);
}

void show_compile_definition(void) 
{
  puts("Compile time definitions:");

#ifdef HOSTNAME
  printf("#define HOSTNAME \"%s\"\n", HOSTNAME);
#endif

#ifdef USE_DOMAIN
  puts("#define USE_DOMAIN");
#endif

#ifdef DOMAIN
  printf("#define DOMAIN \"%s\"\n", DOMAIN);
#endif

#ifdef HIDDEN_HOST
  puts("#define HIDDEN_HOST");
#endif

#ifdef _PGPPATH
  printf("#define PGP %s\n", _PGPPATH);
#endif

#ifdef MIXMASTER_PATH
  puts("#define MIXMASTER_PATH");
#endif

#ifdef HOMESPOOL
  puts("#define HOMESPOOL");
#endif

#ifdef MAILPATH
  printf("#define MAILPATH \"%s\"\n", MAILPATH);
#endif

}

/*
** pms@romance.kaist.ac.kr
*/
static void show_version (void)
{
  printf("\n%s: %s", Mutt_version, Copyright);
#ifdef SHOW_COMPILE_DEF
  show_compile_definition();
#endif
  exit(0);
}

int main (int argc, char **argv)
{
  char folder[_POSIX_PATH_MAX], buffer[LONG_STRING], *subject = 0;
  char *includeFile = 0, *ccList = 0;
  int i, ch, nosysrc = 0, batchmode = 0;
  extern char *optarg;
  extern int optind;

  folder[0] = 0;

  sprintf(buffer, "HanMutt %s", VERSION);
  Mutt_version = safe_strdup(buffer);
  SRAND(time(0));
  setlocale(LC_CTYPE, "");
  umask(077);

  /* Check for a batch send. */
  if (!isatty(0)) batchmode = SENDBATCH;

  /* Process the arguments */
  while ((ch = getopt(argc, argv, "f:c:d:s:i:nvh")) != EOF)
    /* Note: "ch" must be "int" rather than "char" because on some systems,
     * notably AIX, "char" is "unsigned" so the following test is always true
     * since "EOF" is often -1.  */
    switch(ch) {
    case 'f':
      strcpy(folder, optarg);
      break;
    case 'c':
      ccList = optarg;
      break;
    case 'd':
#ifdef DEBUG
      debuglevel = atoi(optarg);
      printf("Debugging at level %d.\n", debuglevel);
#else
      printf("DEBUG was not defined during compilation.  Ignored.\n");
#endif
      break;
    case 'i':
      includeFile = optarg;
      break;
    case 's':
      subject = optarg;
      break;
    case 'v':
      show_version();
    case 'n':
      nosysrc = 1;
      break;
    default:
      mutt_help();
    }

  keymap_init(); /* must come before mutt_init */

  if (mutt_init(nosysrc) == -1)
    ci_any_key_to_continue(0);

  if (!batchmode) {
    initscr();
    keypad(stdscr, TRUE);
    cbreak();
    noecho();

#ifdef NCURSES_MOUSE_VERSION
    if (option(OPTXTERMMOUSE)) {
      Mousemask = mousemask(BUTTON1_CLICKED|BUTTON1_DOUBLE_CLICKED, 0);
      dprint(1, (debugfile, "main(): mousemask = %d.\n", (int)Mousemask));
    }
#endif /* NCURSES_MOUSE_VERSION */

    ci_signal_init();
  }

  if (optind < argc) {
    METAENV *env = mutt_new_metaenv();

    env->env = mutt_new_envelope();
    env->env->subject = safe_strdup(subject);
    for (i=optind; i<argc; i++)
      rfc822_parse_adrlist(&env->env->to, argv[i], "@");
    if (ccList)
      rfc822_parse_adrlist(&env->env->cc, ccList, "@");
    ci_send_message(env, includeFile, batchmode);
    if (!batchmode)
      endwin();
  }
  else {
#ifdef USE_SETGID
    struct group *grp = getgrnam("mail");

    if (!grp) {
      endwin();
      puts("Error!  Couldn't find group entry for \"mail\".");
      exit(-1);
    }
    User_gid = getgid();
    Mail_gid = grp->gr_gid;
    if (setegid(Mail_gid) == -1) {
      dprint(1,(debugfile, "main(): setegid() returned -1; errno = %d.\n", errno));
      endwin();
      puts("Error!  Could not change gid to mail!");
      exit(-1);
    }
#endif /* USE_SETGID */

    if (!folder[0]) strcpy(folder, Spoolfile);
    if ((Context = mutt_open_mailbox(folder, NULL)) == NULL) {
      endwin();
      printf("Folder %s does not exist.\n", folder);
    }
    else
      ci_main();
  }

  exit(0);
}
