/* Copyright (C) 1996 Michael R. Elkins.
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"

#define OFFSET menu->pagestart
#define LENGTH (LINES-4-OFFSET)

MUTTMENU *mutt_newMenu (void)
{
  MUTTMENU *p = (MUTTMENU *)safe_malloc(sizeof(MUTTMENU));

  p->current = p->oldcurrent = p->top = 0;
  p->redraw = REDRAW_FULL;
  p->makeEntry = 0;
  p->max = 0;
  p->title = p->prompt = p->label = 0;
  p->pagestart = 2; /* default value */
#ifdef NCURSES_MOUSE_VERSION
  p->tool = 0;
#endif /* NCURSES_MOUSE_VERSION */
  return(p);
}

void mutt_menuDestroy (MUTTMENU **p)
{
  safe_free((void **)&(*p)->prompt);
  safe_free((void **)&(*p)->title);
  safe_free((void **)&(*p)->label);
  safe_free((void **)p);
}

int mutt_menuLoop (MUTTMENU *menu)
{
  int i;
  char buffer[256];

  FOREVER {
    /*
     * See if all or part of the screen needs to be updated.
     */
    switch (menu->redraw) {
    case REDRAW_FULL:

      clear();
      if (menu->title) CenterLine(menu->title, 0);
      if (menu->label) mvaddstr(1, 0, menu->label);
#ifdef NCURSES_MOUSE_VERSION      
      if (menu->tool) toolbar_post(menu->tool);
#endif /* NCURSES_MOUSE_VERSION */

    case REDRAW_INDEX:

      for (i = menu->top; i < menu->top + LENGTH; i++) {
   if (i < menu->max) {
	  menu->makeEntry(buffer, i);
	  buffer[COLS-3] = 0;
      move(i - menu->top + OFFSET, 0);
      clrtoeol();
	  mvaddstr(i-menu->top+OFFSET, 3, buffer);
	}
	else {
	  move(i-menu->top+OFFSET, 0);
	  clrtoeol();
	}
      }
      menu->redraw = 0;
      break;

    case REDRAW_MOTION:

      mvaddstr(menu->oldcurrent + OFFSET - menu->top, 0, "  ");
      break;
    }

    attron(A_REVERSE);
    mvaddstr(menu->current - menu->top + OFFSET, 0, "->");
    attroff(A_REVERSE);

    refresh();
    i = ci_getch();
    ci_clear_error();

    if (i < KEY_MIN)
      i = keymap_lookup (i, menu->menu);
#ifdef NCURSES_MOUSE_VERSION
    else if (i == KEY_MOUSE) {
      MEVENT m;

      getmouse(&m);
      if (m.y == 0) {
	if ((i = toolbar_select(menu->tool, m.x)) == -1) continue;
      }
      else if (m.y >= OFFSET && m.y <= OFFSET + LENGTH) {
	menu->oldcurrent = menu->current;
	menu->current = m.y - OFFSET + menu->top;
	menu->redraw = REDRAW_MOTION;
	if (m.bstate & BUTTON1_DOUBLE_CLICKED)
	  return(OP_GENERIC_SELECT_ENTRY);
	continue;
      }
      else
	continue;
    }
#endif /* NCURSES_MOUSE_VERSION */
    else
      switch (i) {
      case KEY_LEFT:
      case KEY_PPAGE:

	i = OP_GENERIC_PREVIOUS_PAGE;
	break;

      case KEY_RIGHT:
      case KEY_NPAGE:

	i = OP_GENERIC_NEXT_PAGE;
	break;

      case KEY_UP:

	i = OP_GENERIC_PREVIOUS_ENTRY;
	break;

      case KEY_DOWN:

	i = OP_GENERIC_NEXT_ENTRY;
	break;

      case KEY_HOME:

	i = OP_GENERIC_FIRST_ENTRY;
	break;

      case KEY_END:
	
	i = OP_GENERIC_LAST_ENTRY;
	break;

      default:

	continue;
      }

    switch (i) {
    case OP_GENERIC_NEXT_ENTRY:

      if (menu->current < menu->max - 1) {
	menu->oldcurrent = menu->current;
    menu->current++;
    if (menu->current >= menu->top + LENGTH) {
	  menu->top = menu->current;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else {
	mutt_error("You are on the last entry.");
      }
      break;

    case OP_GENERIC_PREVIOUS_ENTRY:

      if (menu->current > 0) {
	menu->oldcurrent = menu->current;
	if (--menu->current < menu->top) {
      if ((menu->top = menu->current - LENGTH + 1) < 0) menu->top = 0;
	  menu->redraw = REDRAW_INDEX;
	}
	else
	  menu->redraw = REDRAW_MOTION;
      }
      else {
	mutt_error("You are on the first entry.");
      }
      break;
	
    case OP_GENERIC_NEXT_PAGE:

      if (menu->top + LENGTH <= menu->max - 1) {
	menu->top += LENGTH;
	menu->current = menu->top;
	menu->redraw = REDRAW_INDEX;
      }
      else {
	mutt_error("You are on the last page.");
      }
      break;

    case OP_GENERIC_PREVIOUS_PAGE:

      if (menu->top > 0) {
	if ((menu->top -= LENGTH) < 0) menu->top = 0;
	menu->current = menu->top + LENGTH - 1;
	menu->redraw = REDRAW_INDEX;
      }
      else {
	mutt_error("You are on the first page.");
      }
      break;

    case OP_GENERIC_FIRST_ENTRY:

      menu->oldcurrent = menu->current;
      if (menu->top > 0)
	menu->redraw = REDRAW_INDEX;
      else
	menu->redraw = REDRAW_MOTION;
      menu->top = menu->current = 0;
      break;

    case OP_GENERIC_LAST_ENTRY:

      menu->oldcurrent = menu->current;
      menu->current = menu->max - 1;
      if (menu->current > menu->top + LENGTH - 1) {
	menu->top = menu->current - LENGTH;
	if (menu->top < 0) menu->top = 0;
	menu->redraw = REDRAW_INDEX;
      }
      else
	menu->redraw = REDRAW_MOTION;
      break;

    case OP_GENERIC_REFRESH:

      menu->redraw = REDRAW_FULL;
      break;

    case OP_GENERIC_HELP:

      if (help(menu->menu)) menu->redraw = REDRAW_FULL;
      break;

    case OP_NULL:

      beep();
      mutt_error("Key is not bound.  Press '?' for help.");
      break;

    default:
      
      return (i);
    }
  }
  /* not reached */
}
