/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <stdlib.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"

#ifdef HAVE_LIBGEN
#include <libgen.h>
#endif

#ifdef HAVE_REGEX_H
#include <regex.h>
#endif

int ci_simple_pager (const char *fname)
{
  extern int IsPrint(int);
  char buffer[LONG_STRING];
  long **offset = 0, offset_max = 0, offset_len = 0;
  int i, ch = 0, rc = -1, chars = 0, curline = 0, lines = 0, y, x;
  unsigned char *pending = 0;
  struct stat sb;

#ifdef HAVE_REGEX
  char *re=0;
#endif /* HAVE_REGEX */

#if defined(HAVE_REGEX) || defined(HAVE_RE_EXEC)
  char searchbuf[STRING];
  int searching=0, savecurline=0;
#endif

  FILE *fp = fopen(fname, "r");

  if (!fp) {
    mutt_error("Could not open file %s!", fname);
    return(-1);
  }
  scrollok(stdscr, TRUE);
  stat(fname, &sb);
  unlink(fname);
  clear();
  while (ch != -1) {
    while (lines < LINES-1) {
      if (!pending) {
	if (offset_max == offset_len) {
	  if (offset == 0)
	    offset = safe_malloc(sizeof(long *)*((offset_max = LINES)));
	  else
	    safe_realloc((void **)&offset,
			 sizeof(long *) * ((offset_max += LINES)));
	}
	if (curline == offset_len) {
	  offset[curline] = safe_malloc(sizeof(long));
	  *offset[curline] = ftell(fp);
	  offset_len++;
	}
	if (fgets(buffer, sizeof(buffer), fp) == NULL) break;	      
	pending = buffer;
      }

#ifdef HAVE_REGEX
#define REGEX(A,B) regex(A,B)
#elif HAVE_RE_EXEC
#define REGEX(A,B) re_exec(B)
#endif

#if defined(HAVE_REGEX) || defined(HAVE_RE_EXEC)
      if (searching) {
	if (REGEX(re, buffer) != 0) {
	  clear();
	  searching = 0;
	  curline -= 3;
	  if (curline < 0) curline = 0;
	  fseek(fp, *offset[curline], 0);
	  continue;
	}
	pending = 0;
	curline++;
	continue;
      }
#endif

      chars = 0;
      while (*pending && lines<LINES-1) {
	if (chars == COLS) {
	  lines++;
	  chars = 0;
	  /* If the last character on the line is right at the edge of
	   * the screen, don't print the newline, or else we'll end up
	   *  with a blank line.
	   */
	  if (*pending == '\n') pending++;
	  continue;
	}
	if (*pending == '\n') {
	  addch(*pending);
	  chars = 0;
	  lines++;
	}
#define MUTT_TABSTOP 8
#define next_tabstop (chars > (COLS/8)*8 ? 0 : (((chars/MUTT_TABSTOP)+1)*MUTT_TABSTOP))
	else if (*pending == '\t') {
	  i = next_tabstop;
	  if (i == 0) {
	    addch('\n');
	    lines++;
	    chars = 0;
	  }
	  else
	    for (i=i-chars; i > 0; i--, chars++) addch(' ');
	}
/* PMS */
	else if (IsPrint(*pending)) {
	  addch(*pending);
	  chars++;
	}
	else if (iscntrl(*pending)) {
	  if (chars > COLS-1) {
	    addch('\n');
	    chars = 0;
	    lines++;
	    continue;
	  }
	  addch('^');
	  addch(*pending+'A'-1);
	  chars += 2;
	  if (*pending == ctrl('L')) {
	    pending++;
	    break;
	  }
	}
	else
	  addch('?'); /* unprintable character! */
	pending++;
      }

      if (*pending == 0) {
	/* finished with this line */
	curline++;
	pending = 0;
      }
    }

#if defined(HAVE_REGEX) || defined(HAVE_RE_EXEC)
    if (searching) {
      move(LINES-1, 0);
      clrtoeol();
      attrset(A_REVERSE);
      addstr("Not found!");
      attrset(A_NORMAL);
      searching=0;
      curline=savecurline;
      fseek(fp, *offset[curline-1], 0);
    }
    else
#endif
    {
      if (!option(OPTFLOATINGPROMPT)) {
   move(LINES-1, 0);
      }

      attron(A_REVERSE);
      if (ftell(fp) < sb.st_size-1) {
#ifdef DEBUG
	int y, x;
	getyx(stdscr, y, x);
	dprint(10, (debugfile, "simple_pager: prompt location = (%d,%d)\n",
		    x, y));
#endif
	printw("More (you've seen %d%%)", 100*ftell(fp)/sb.st_size);
      }
      else {
	addstr("(END)");
      }
      attroff(A_REVERSE);
    }

    refresh();
    ch = ci_getch();

#ifdef NCURSES_MOUSE_VERSION
    if (ch == KEY_MOUSE) {
      MEVENT e;

      getmouse(&e);
      if (e.bstate & BUTTON1_CLICKED) {
	if (e.y <= LINES/2)
	  ch = OP_PAGER_NEXT_PAGE;
	else
	  ch = OP_PAGER_PREVIOUS_PAGE;
      }
      else
	continue;
    }
    else
#endif
      ch = keymap_lookup(ch, MENU_PAGER);

    rc = ch;

    switch (ch) {
    case OP_PAGER_EXIT:
      rc = -1;
      ch = -1;
      break;

    case OP_PAGER_NEXT_PAGE:
      if (ftell(fp) < sb.st_size-1) {
	lines = 0;
	clear();
      }
      else {
	/* end of the current message, so display the next message. */
	rc = OP_MAIN_NEXT_UNDELETED;
	ch = -1;
      }
      break;

    case OP_PAGER_PREVIOUS_PAGE:
      if (curline != 0) {
	clear();
	curline -= 2*LINES-2;
	if (curline < 0) curline = 0;
	fseek(fp, *offset[curline], 0);
	lines = 0;
      }
      break;

    case OP_PAGER_NEXT_LINE:
      getyx(stdscr, y, x);
      dprint(10, (debugfile, "simple_pager: location before clrtoeol() = (%d,%d)\n", x, y));
      move(y, 0);
      clrtoeol();
      if (ftell(fp) < sb.st_size-1)
	lines--;
      else
	beep();
      break;

    case OP_PAGER_PREVIOUS_LINE:
      clear();
      if ((curline -= LINES + 1) < 0) curline = 0;
      fseek(fp, *offset[curline], 0);
      lines = 0;
      break;

    case OP_PAGER_TOP:
      if (curline > LINES) {
	clear();
	lines=0;
	fseek(fp, 0, 0);
	curline=0;
      }
      else
	beep();
      break;

#if defined(HAVE_REGEX) || defined(HAVE_RE_EXEC)
    case OP_PAGER_SEARCH:
      move(LINES-1, 0);
      clrtoeol();
      addstr("Search: ");
      searchbuf[0] = 0;
      ci_enter_string(searchbuf, LINES-1, sizeof("Search: ")-1, 0, 0);

#ifdef HAVE_REGEX
      re = regcmp(searchbuf, 0);
#elif HAVE_RE_EXEC
      re_comp(searchbuf);
#endif

      searching=1;
      lines=0;
      savecurline = curline;
      curline -= LINES-2;
      if (curline < 0) curline=0;
      fseek(fp, *offset[curline], 0);
      break;
#endif

    default:
      ch = -1;
    }
  }
  fclose(fp);
  for (i=0;i<offset_len; i++) safe_free((void **)&offset[i]);
  safe_free((void **)&offset);
  return (rc != -1 ? rc : 0);
}
