/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include "muttlib.h"
#include "rfc822.h"
#include "rfc1522.h"

void mutt_parse_references (LIST **l, char *s)
{
  LIST *t, *r = *l;

  while (r && r->next) r = r->next;
  while ((s = strtok(s, " \t")) != NULL) {
    t = (LIST *)safe_malloc(sizeof(LIST));
    t->data = safe_strdup(s);
    t->next = NULL;
    if (*l) {
      r->next = t;
      r = r->next;
    }
    else
      *l = r = t;
    s = NULL;
  }
}

static int check_encoding (const char *c)
{
  if (strncasecmp("7bit", c, sizeof("7bit")-1) == 0)
    return(ENC7BIT);
  else if (strncasecmp("8bit", c, sizeof("8bit")-1) == 0)
    return(ENC8BIT);
  else if (strncasecmp("binary", c, sizeof("binary")-1) == 0)
    return(ENCBINARY);
  else if (strncasecmp("quoted-printable", c, sizeof("quoted-printable")-1) == 0)
    return(ENCQUOTEDPRINTABLE);
  else if (strncasecmp("base64", c, sizeof("base64")-1) == 0)
    return(ENCBASE64);
  else
    return(ENCOTHER);
}

static PARAMETER *parse_parameters (char *s)
{
  PARAMETER *head = 0, *cur = 0, *new;
  char buffer[STRING], *p, savechar;
  short i;

  while (*s) {
    p = strchr(s, '=');
    if (!p) {
      dprint(1, (debugfile, "parse_parameters: malformed parameter: %s\n", s));
      return(head); /* just bail out now */
    }
    savechar = *p;
    *p = 0;
    new = mutt_new_parameter();
    new->attribute = safe_strdup(s);
    *p = savechar;
    p++;
    if (*p == '"') {
      p++;
      for (i=0; *p != '"'; i++, p++) {
        if (*p == '\\') {
          /* Quote the next character */
          buffer[i] = *(p+1);
          p++;
        }
        else
          buffer[i] = *p;
      }
      buffer[i] = 0;
    }
    else {
      for (i=0; *p && *p != ' ' && *p != ';'; i++, p++) {
        buffer[i] = *p;
      }
      buffer[i] = 0;
    }
    new->value = safe_strdup(buffer);

    /* Add this parameter to the list */
    if (head) {
      cur->next = new;
      cur = cur->next;
    }
    else
      head = cur = new;

    /* Find the next parameter */
    if (*p == ';') p++;
    else {
      while (*p && *p != ';') p++;
      if (!*p) break; /* no more parameters! */
      p++;
    }

    /* Move past any leading whitespace */
    while (*p && whitespace(*p))
      p++;
    s = p;
  }    
  return(head);
}

static void parse_content_type (char *s, BODY *ct)
{
  char *pc;
  char buffer[SHORT_STRING];
  short i = 0;

  /* First extract any existing parameters */
  pc = strchr(s, ';');
  if (pc) {
    *pc = 0;
    pc++;
    while (*pc && whitespace(*pc)) pc++;
    ct->parameter = parse_parameters(pc);
  }
  
  /* Now get the subtype */
  pc = strchr(s, '/');
  if (pc) {
    *pc = 0;
    pc++;
    while (*pc && !whitespace(*pc) && *pc != ';') {
      buffer[i++] = *pc;
      pc++;
    }
    buffer[i] = 0;
    if (ct->subtype)
      safe_free((void **)&ct->subtype);
    ct->subtype = safe_strdup(buffer);
  }

  /* Finally, get the major type */
  if (strcasecmp("text", s) == 0)
    ct->type = TYPETEXT;
  else if (strcasecmp("image", s) == 0)
    ct->type = TYPEIMAGE;
  else if (strcasecmp("audio", s) == 0)
    ct->type = TYPEAUDIO;
  else if (strcasecmp("video", s) == 0)
    ct->type = TYPEVIDEO;
  else if (strcasecmp("multipart", s) == 0)
    ct->type = TYPEMULTIPART;
  else if (strcasecmp("application", s) == 0)
    ct->type = TYPEAPPLICATION;
  else if (strcasecmp("message", s) == 0)
    ct->type = TYPEMESSAGE;
  else
    ct->type = TYPEOTHER;

  if (ct->subtype == 0) {
    /* Some older non-BODY mailers (i.e., mailtool, elm) have a content-type
     * field, so we can attempt to convert the type to BODY here.
     */
    if (ct->type == TYPETEXT)
      ct->subtype = safe_strdup("plain");
    else if (ct->type == TYPEAUDIO)
      ct->subtype = safe_strdup("basic");
    else if (ct->type == TYPEMESSAGE)
      ct->subtype = safe_strdup("rfc822");
    else if (ct->type == TYPEOTHER) {
      ct->type = TYPEAPPLICATION;
      sprintf(buffer, "x-%s", s);
      ct->subtype = safe_strdup(buffer);
    }
  }
}

BODY *mutt_read_mime_header (FILE *fp) {
  BODY *p = mutt_new_body();
  short linelen = 0;
  char *c;
  char buffer[LONG_STRING], line[LONG_STRING];
  PARAMETER *parms;
  
  line[0] = 0;
  p->hdr_offset = ftell(fp);
  p->encoding = ENC7BIT; /* default from RFC1521 */
  p->type = TYPETEXT;
  while (fgets(buffer, LONG_STRING, fp) != NULL) {

    if (whitespace(buffer[0]) && buffer[0] != '\n' &&
        ! (buffer[0] == '\r' && buffer[1] == '\n')) {
      /* Remove the trailing whitespace from line */
      c = buffer + strlen(buffer) - 1;
      while (whitespace(*c) && c >= buffer) *c-- = 0;

      /* Convert whitespace to a single space */
      c = buffer;
      while (whitespace(*c)) c++;
      sprintf(line+linelen, " %s", c);
      linelen += strlen(c) + 1;
    }
    else if (line[0]) {

      /* Find the value of the current header */
      c = strchr(line, ':');
      if (c) {
        c++;
        while (*c && whitespace(*c)) c++;
      }
      else {
        dprint(1, (debugfile, "read_mime_header: bogus BODY header: %s\n", line));
        break;
      }

      switch(tolower(line[0])) {
      case 'c':
        if (strncasecmp("ontent-type", line+1, sizeof("ontent-type")-1) == 0)
          parse_content_type(c, p);
        else if (strncasecmp("ontent-transfer-encoding", line+1, sizeof("ontent-transfer-encoding")-1) == 0)
          p->encoding = check_encoding(c);
        else if (strncasecmp("ontent-disposition", line+1, sizeof("ontent-disposition")-1) == 0) {
          if (strncasecmp("attachment", c, sizeof("attachment")-1) == 0)
            p->disposition = DISPATTACH;
          else
            p->disposition = DISPINLINE;

          /* Check to see if a default filename was given */
          if ((c = strchr(c, ';')) != NULL) {
            c++;
            while (*c && whitespace(*c)) c++;
            if ((c = mutt_get_parameter("filename", (parms = parse_parameters(c)))) != 0)
              p->filename = safe_strdup(c); 
            mutt_free_parameter(&parms);
          }
        }
        else if (strncasecmp("ontent-description", line+1, sizeof("ontent-description")-1) == 0) {
          if (!p->description) p->description = safe_strdup(c);
        }
        else if (strncasecmp("ontent-md5", line+1, sizeof("ontent-md5")-1) == 0) {
          p->md5 = safe_strdup(c);
        }
        break;
      }

      /* Check for end of headers */
      if (buffer[0] == '\n' || (buffer[0] == '\r' && buffer[1] == '\n'))
        break;

      linelen = strlen(buffer);
      c = buffer+linelen-1;
      while (whitespace(*c) && c >= buffer) {
        *c-- = 0;
        linelen--;
      }
      strcpy(line, buffer);
    }
    else if (buffer[0] == '\n' || (buffer[0] == '\r' && buffer[1] == '\n'))
      break; /* done */
    else {
      linelen = strlen(buffer);
      c = buffer+linelen-1;
      while (whitespace(*c) && c >= buffer) {
        *c-- = 0;
        linelen--;
      }
      strcpy(line, buffer);
    }
  }
  p->offset = ftell(fp); /* Mark the start of the real data */
  if (p->type == TYPETEXT && !p->subtype) p->subtype = safe_strdup("plain");
  return(p);
}

/* Routine to parse a MESSAGE/RFC822 body */
BODY *parse_messageRFC822 (FILE *fp, BODY *parent)
{
  BODY *msg;

  msg = mutt_read_mime_header(fp);
  if (msg->type == TYPEMULTIPART)
    msg->parts = parse_multipart(fp, mutt_get_parameter("boundary", msg->parameter));
  else if (msg->type == TYPEMESSAGE)
    msg->parts = parse_messageRFC822(fp, msg);
  else
    msg->length = parent->length - (msg->offset - parent->offset);
  return(msg);
}

BODY *parse_multipart (FILE *fp, char *boundary)
{
  int blen = strlen(boundary), len, crlf = 0;
  char buffer[LONG_STRING];
  BODY *head = 0, *last = 0, *new = 0;
  short i;

  while (fgets(buffer, LONG_STRING, fp) != NULL) {
    len = strlen(buffer);

    /* take note of the line ending.  I'm assuming that either all endings
     * will use <CR><LF> or none will.
     */
    if (buffer[len-2] == '\r') crlf = 1;

    if (buffer[0] == '-' && buffer[1] == '-' && strncmp(buffer+2, boundary, blen) == 0) {
      if (last) {
        last->length = ftell(fp) - last->offset - len - 1 - crlf;
        if (last->parts && last->parts->length == 0)
          last->parts->length = ftell(fp) - last->parts->offset - len - 1 - crlf;
      }
      /* Remove any trailing whitespace, up to the length of the boundary */
      for(i=len-1; whitespace(buffer[i]) && i>blen+2; i--)
        buffer[i] = 0;
      /* Check for the end boundary */
      if (buffer[2+blen] == '-' && buffer[3+blen] == '-' && buffer[4+blen] == 0)
        break;

      new = mutt_read_mime_header(fp);
      if (head) {
        last->next = new;
        last = new;
      }
      else
        last = head = new;
      
      if (new->type == TYPEMULTIPART)
        new->parts = parse_multipart (fp, mutt_get_parameter("boundary", new->parameter));
      else if (new->type == TYPEMESSAGE)
        new->parts = parse_messageRFC822 (fp, new);
    }
  }
  return(head);
}

static void parse_date (char *s, HEADER *h)
{
  int count = 0;
  char *t = strchr(s, ',');
  int year, hour, min, sec;

  h->zoccident = 0;

  /* kill the day of the week, if it exists. */
  if (t) {
    t++;
    while (*t && isspace(*t)) t++;
  }
  else
    t = s;

  while ((t = strtok(t, " \t")) != NULL) {
    switch (count) {
    case 0: /* day of the month */
      h->day = atoi (t);
      break;
    case 1: /* month of the year */
      switch(tolower(t[0])) {
      case 'j':
        switch (tolower(t[1])) {
        case 'a':
          h->month = 0;
          break;
        default:
          if (tolower(t[2]) == 'n')
            h->month = 5;
          else
            h->month = 6;
        }
        break;
      case 'f':
        h->month = 1;
        break;
      case 'm':
        if (tolower(t[2]) == 'r') h->month = 2;
        else h->month = 4; /* May */
        break;
      case 'a':
        if (tolower(t[1]) == 'p') h->month = 3; /* April */
        else h->month = 7; /* August */
        break;
      case 's':
        h->month = 8; /* September */
        break;
      case 'o':
        h->month = 9; /* October */
        break;
      case 'n':
        h->month = 10; /* November */
        break;
      case 'd':
        h->month = 11; /* December */
        break;
      default:
	h->month = 12; /* ERROR! */
      }
      break;
    case 2: /* year */
      year = atoi(t);
      if (year < 1900) year += 1900;
      h->year = year - 1971;
      break;
    case 3: /* time of day */
      if (sscanf(t, "%d:%d:%d", &hour, &min, &sec) == 3);
      else if (sscanf(t, "%d:%d", &hour, &min) == 2) {
        sec = 0;
      }
      else {
        dprint(1, (debugfile, "parse_date: could not process time format: %s\n", t));
        hour = min = sec = 0;
      }
      h->hours = hour;
      h->minutes = min;
      h->seconds = sec;
      break;
    case 4: /* timezone */
      if (*t == '+' || *t == '-') {
        if (*t == '-')
          h->zoccident = 1;
        t++;
        h->zhours = (t[0] - '0') * 10 + (t[1] - '0');
        h->zminutes = (t[2] - '0') * 10 + (t[3] - '0');
      }
      else {
        /* Most all mail should use the +HHMM format instead of character
         * strings, but here are a few defaults for the US...  */
        if (*t == 'C') {
          h->zhours = 6;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        } else if (*t == 'E') {
          h->zhours = 5;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
        else if (*t == 'M') {
          h->zhours = 7;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
        else if (*t == 'P') {
          h->zhours = 8;
          h->zoccident = 1;
          if (t[1] == 'D') h->zhours--;
        }
      }
    }
    count++;
    t = 0;
  }
}

/* This is one place where checks for buffer overflow needs to take place.
 * Currently, it tries to gather up all of the continuation lines for a
 * header before processing it.  However, this creates problems for
 * messages where there are a huge number of listed recipients...
 */
#define HUGE_STRING 5120
static void read_rfc822_header (FILE *fp, HEADER *hdr)
{
  int linelen = 0;
  char buffer[HUGE_STRING];
  char line[HUGE_STRING];
  char *val;
  long content_length = -1;

  if (!hdr->content) {
    hdr->content = mutt_new_body();
    hdr->content->type = TYPETEXT;
    hdr->content->subtype = safe_strdup("Plain");
    hdr->content->encoding = ENC7BIT;
  }

  line[0] = 0;
  while (fgets(buffer, HUGE_STRING, fp) != NULL) {
    /* Remove the trailing whitespace */
    val = buffer + strlen(buffer) - 1;
    while (whitespace(*val) && val >= buffer) {
      *val = 0;
      val--;
    }

    if (buffer[0] && whitespace(buffer[0])) {
      /* Convert leading whitespace to a single space */
      val = buffer;
      while (*val && whitespace(*val)) val++;
/* Simply merge the next line - pms@romance.kaist.ac.kr */
      sprintf(line+linelen, "%s", val);
      linelen += strlen(val);
    }
    else if (line[0]) {
      val = strchr(line, ':');
      if (val) {
        val++;
        while (*val && whitespace(*val)) val++;
      }
      switch(tolower(line[0])) {
      case 'a':
	if (strncasecmp("pparently-", line+1, 10) == 0) {
          if (strncasecmp("to:", line+11, 3) == 0)
            rfc822_parse_adrlist(&hdr->to, val, Hostname);
          else if (strncasecmp("from:", line+11, 5) == 0)
            rfc822_parse_adrlist(&hdr->from, val, Hostname);
	}
	break;
      case 'c':
        if (strncasecmp("ontent-", line+1, sizeof("ontent-")-1) == 0) {
          if (strncasecmp("length:", line+sizeof("ontent-"), sizeof("length:")-1)==0) {
            content_length = atoi(val);
            dprint(1,(debugfile,"read_rfc822_header: honoring Content-Length = %ld in message %d.\n", content_length, hdr->msgno));
          }
          else if (strncasecmp("transfer-encoding:", line+sizeof("ontent-"), sizeof("transfer-encoding:")-1) == 0)
            hdr->content->encoding = check_encoding(val);
          else if (strncasecmp("type:", line+sizeof("ontent-"), sizeof("type:")-1) == 0)
            parse_content_type(val, hdr->content);
          else if (strncasecmp("md5:", line+sizeof("ontent-"), sizeof("md5:")-1) == 0)
            hdr->content->md5 = safe_strdup(val);
        }
        else if (strncasecmp("c:", line+1, sizeof("c:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->cc, val, Hostname);
          rfc1522_decode_adrlist(hdr->to);
        }
        break;
      case 'd':
        if (strncasecmp("date:", line, sizeof("date:")-1) == 0)
          parse_date(val, hdr);
        break;
      case 'f':
        if (strncasecmp("from:", line, sizeof("from:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->from, val, Hostname);
          rfc1522_decode_adrlist(hdr->from);
        }
        break;
      case 'm':
        if (strncasecmp("ime-Version:", line+1, sizeof("ime-Version:")-1) == 0)
          hdr->mime = 1;
        else if (strncasecmp("essage-ID:", line+1, sizeof("essage-ID:")-1) == 0)
          hdr->message_id = safe_strdup(val);
        break;
      case 'r':
        if (strncasecmp("eply-to:", line+1, sizeof("eply-to:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->reply_to, val, Hostname);
          rfc1522_decode_adrlist(hdr->reply_to);
        }
        else if (strncasecmp("eferences:", line+1, 10) == 0)
          mutt_parse_references(&hdr->references, val);
        break;
      case 's':
        if (strncasecmp("tatus:", line+1, sizeof("tatus:")-1) == 0) {
          while (*val) {
            switch(*val){
            case 'r':
              hdr->replied = 1;
              break;
            case 'O':
              hdr->old = 1;
              break;
            case 'R':
              hdr->read = 1;
              break;
            }
            val++;
          }
        }
        else if (strncasecmp("ubject:", line+1, sizeof("ubject:")-1) == 0) {
          if (!hdr->subject) {
            hdr->subject = safe_strdup(val);
            rfc1522_decode(hdr->subject, strlen(hdr->subject), 0);
	    mutt_tabs_to_spaces(hdr->subject);
          }
        }
        break;
      case 't':
        if (strncasecmp("o:", line+1, sizeof("o:")-1) == 0) {
          rfc822_parse_adrlist(&hdr->to, val, Hostname);
          rfc1522_decode_adrlist(hdr->to);
        }
        break;
      case 'x':
        /* Be compatible with C-CLIENT so we don't lose any flags */
        if (strncasecmp("-status:", line+1, sizeof("-status:")-1) == 0) {
          while (*val) {
            switch(*val){
            case 'A':
              hdr->replied = 1;
              break;
            case 'D':
              hdr->deleted = 1;
              break;
	    case 'F':
	      hdr->flagged = 1;
            }
            val++;
          }
        }
        break;
      }

      /* Check for end of header. */
      if (buffer[0] == 0 || (!whitespace(buffer[0]) && strchr(buffer, ':') == 0))
        break;

      strcpy(line, buffer);
      linelen = strlen(buffer);
    }
    else {
      strcpy(line, buffer);
      linelen = strlen(buffer);
    }
  }
  hdr->content->hdr_offset = hdr->offset;
  hdr->content->offset = ftell(fp);

  if (option(OPTHONORLENGTH) && content_length > 0) {
    /* If a Content-Length field was given, honor it. */
    fseek(fp, content_length, 1);
  }
}

static int mutt_first_message (CONTEXT *ctx)
{
  int i, old = -1;

  for (i = 0; i < ctx->vcount; i++) {
    if (! ctx->hdrs[ *ctx->v2r[i] ]->read) {
      if (! ctx->hdrs[ *ctx->v2r[i] ]->old)
        return(i);
      else if (old < 0)
        old = i;
    }
  }
  if (old < 0) {
    /*
     * There were no old messages, so position on the last message received.
     */
    if (Sort & SORTREVERSE)
      old = 0;
    else
      old = ctx->vcount - 1;
  }
  return(old);
}

/* Note that this function is also called when new mail is appended to the
 * currently open folder, and NOT just when the mailbox is initially read.
 *
 * MMDF folders look just like MBOX folders, except that they have some
 * extra garbage around them.  Luckily we can just ignore that...
 *
 * NOTE: it is assumed that the mailbox being read has been locked before
 * this routine gets called.  Strange things could happen if it's not!
 */
int mutt_parse_mailbox (CONTEXT *ctx)
{
  struct stat sb;
  int i, oldmax = ctx->hdrmax, oldcount = ctx->msgcount, count = 0, lines = 0;
  char buf[HUGE_STRING];
  HEADER *curhdr;
  time_t t;

  /* Save information about the folder at the time we opened it. */
  stat(ctx->folder->name, &sb);
  ctx->folder->size = sb.st_size;
  ctx->folder->mtime = sb.st_mtime;

  dprint(1,(debugfile, "mutt_parse_mailbox(): size=%d, mtime=%d\n",
	    (int)sb.st_size, (int)sb.st_mtime));

  while (fgets(buf, LONG_STRING, ctx->folder->fp) != NULL) {
    if ((t = is_from(buf))) {
      long loc = ftell(ctx->folder->fp) - strlen(buf);

      /* Save the Content-Length of the previous message */
      if (count > 0) {
#define PREVIOUS ctx->hdrs[ctx->msgcount-1]

        PREVIOUS->length = loc - PREVIOUS->offset - 1;
        PREVIOUS->content->length = loc - PREVIOUS->content->offset - 1;
        PREVIOUS->lines = lines - 1;
#undef PREVIOUS
      }

      if (ctx->msgcount == ctx->hdrmax) {
        /* Need to allocate more memory for structure. */
        if (ctx->hdrs)
          safe_realloc((void **)&ctx->hdrs, sizeof(HEADER *) * ((ctx->hdrmax += 25)));
        else {
          ctx->hdrs = (HEADER **)safe_malloc(sizeof(HEADER *) * ((ctx->hdrmax = 25)));
          for (i = ctx->msgcount; i < ctx->hdrmax; i++)
            ctx->hdrs[i] = 0;
        }
      }
      
      curhdr = ctx->hdrs[ctx->msgcount] = mutt_new_header();

      curhdr->received = t;
      dprint(3, (debugfile, "parse_mailbox(): message %d received at %ld.\n",
                 ctx->msgcount, (long)curhdr->received));

      curhdr->offset = loc;
      curhdr->msgno = ctx->msgcount+1;
      read_rfc822_header(ctx->folder->fp, curhdr);
#ifdef _PGPPATH
      /* NOTE: this _must_ be done before the check for metamail! */
      curhdr->pgp = pgp_query(curhdr->content);
#endif /* _PGPPATH */
#ifdef METAMAIL
#ifdef _PGPPATH
      if (!curhdr->pgp)
#endif /* _PGPPATH */
        if (mutt_needs_metamail(curhdr->content)) curhdr->metamail = 1;
#endif /* METAMAIL */
      if (!curhdr->read && !curhdr->old) ctx->new++;
      ctx->msgcount++;
      count++;

      lines = 0;
    }
    else
      lines++;
  }
  
  /* Only set the content-length of the previous message if we have read more
   * than one message during _this_ invocation.  If this routine is called
   * when new mail is received, we need to make sure not to clobber what
   * previously was the last message since the headers may be sorted.
   */
  if (count > 0) {
    ctx->hdrs[ctx->msgcount-1]->length = ftell(ctx->folder->fp) - ctx->hdrs[ctx->msgcount-1]->offset - 1;
    ctx->hdrs[ctx->msgcount-1]->content->length = ftell(ctx->folder->fp) - ctx->hdrs[ctx->msgcount-1]->content->offset - 1;
    ctx->hdrs[ctx->msgcount-1]->lines = lines - 1;
  }
  
  if (oldmax) {
    /* Need to allocate more space for the virtual=>real mapping. */
    if (oldmax != ctx->hdrmax) {
      safe_realloc((void **)&ctx->v2r, sizeof(short *) * ctx->hdrmax);
      for (i = oldcount; i < ctx->hdrmax; i++) {
        ctx->v2r[i] = (short *)safe_malloc(sizeof(short));
        *ctx->v2r[i] = -1;
      }
    }
  }
  else if (count > 0) {
    ctx->v2r = (short **)safe_malloc(sizeof(short *) * ctx->hdrmax);
    for (i = 0; i < ctx->hdrmax; i++) {
      ctx->v2r[i] = (short *)safe_malloc(sizeof(short));
      *ctx->v2r[i] = -1;
    }
  }

  mutt_sort_headers(ctx);

  if (!oldmax) {
    ctx->current = option(OPTPOINTNEW) ? mutt_first_message(ctx) : 0;
  }

  return(0);
}

void mutt_parse_mime_message (HEADER *cur)
{
  if (cur->content->type != TYPEMESSAGE && cur->content->type != TYPEMULTIPART)
    return; /* nothing to do */
  fseek(Context->folder->fp, cur->content->offset, 0);
  if (cur->content->type == TYPEMULTIPART) {
    if (!cur->content->parts) cur->content->parts = parse_multipart(Context->folder->fp, mutt_get_parameter("boundary", cur->content->parameter));
  }
  else {
    if (!cur->content->parts)
      cur->content->parts = parse_messageRFC822(Context->folder->fp, cur->content);
  }
}

