/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

/* Attachment menu for the curses interface. */

#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <string.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "keymap.h"
#include "mutt_menu.h"

struct attachptr {
  BODY *content;
  struct attachptr *next;
};

struct attachptr **idx = 0;

static struct attachptr *gen_attach_list (BODY *m)
{
  struct attachptr *top = 0, *end = 0, *new;

  while (m) {
    if (m->type == TYPEMULTIPART) {
      if (top)
        end->next = gen_attach_list(m->parts);
      else
        end = top = gen_attach_list(m->parts);
      while (end->next) end = end->next;
    }
    else {
      new = (struct attachptr *)safe_malloc(sizeof(struct attachptr));
      new->next = 0;
      new->content = m;

      if (top) {
        end->next = new;
        end = new;
      }
      else
        top = end = new;
    }
    m = m->next;
  }
  return(top);
}


static int can_print (BODY *m)
{
  if (m->type == TYPETEXT) {
    if (strcasecmp("plain", m->subtype) == 0) return 1;
  }
  else if (m->type == TYPEAPPLICATION) {
    if (strcasecmp("postscript", m->subtype) == 0) return 1;
  }
  return 0;
}

static void display_attachment (BODY *a)
{
  STATE *s;
  char *tempfile = mutt_new_tempfile();

#ifdef METAMAIL
  if (mutt_needs_metamail(a)) {
    FILE *fpout;
    char buffer[LONG_STRING];

    fpout = fopen(tempfile, "w");
    fseek(MAILFP, a->offset, 0);
    mutt_copy_bytes(MAILFP, fpout, a->length);
    fclose(fpout);
    sprintf(buffer, "%s -b -c %s/%s -P -z %s", METAMAIL, TYPE(a->type),
                    a->subtype, tempfile);
    mutt_system(buffer);
    safe_free((void **)&tempfile);
    return;
  }
#endif

  s = mutt_new_state();
  s->fpout = fopen(tempfile, "w");
  s->fpin = MAILFP;
  s->displaying = 1;
  mutt_body_handler(a, s);
  fclose(s->fpout);
  mutt_free_state(&s);
  ci_simple_pager(tempfile);
  safe_free((void **)&tempfile);
}

void attach_makeEntry (char *b, int num)
{
  char t[SHORT_STRING];
  BODY *m = idx[num]->content;

  sprintf(t, "%s/%s", TYPE(m->type), m->subtype);
  sprintf(b, " %2d %-20.20s  %-6.6s (%6ld) %s", num+1, t, ENCODING(m->encoding), m->length, m->description ? m->description : "");
}

void ci_attach (void)
{
  BODY *cur = Context->hdrs[*Context->v2r[Context->current]]->content;
  struct attachptr *list, *plist;
  int idxlen = 0, count = 0;
  char buffer[STRING];
  MUTTMENU *menu;

  plist = list = gen_attach_list(cur);
  for (count=0; plist; count++, plist = plist->next) {
    if (count == idxlen) {
      if (idxlen == 0)
        idx = safe_malloc(sizeof(struct attachptr *) * ((idxlen = 5)));
      else
        safe_realloc((void **)&idx, ((idxlen += 5)) * sizeof(struct attachptr *));
    }
    idx[count] = plist;
  }

  menu = mutt_newMenu();
  sprintf(buffer, "Attachment Menu (%d attachment%s)", count, count != 1 ? "s" : "");
  menu->title = safe_strdup(buffer);
  menu->max = count;
  menu->makeEntry = attach_makeEntry;
  menu->menu = MENU_ATTACH;
#ifdef NCURSES_MOUSE_VERSION
  if (Mousemask) {
    menu->tool = Generic_toolbar;
    toolbar_init(Generic_toolbar);
  }
#endif /* NCURSES_MOUSE_VERSION */

  FOREVER {
    switch (mutt_menuLoop(menu)) {
    case OP_GENERIC_SELECT_ENTRY:

      display_attachment(idx[menu->current]->content);
      menu->redraw = REDRAW_FULL;
      break;

    case OP_ATTACH_PRINT_ENTRY:

      addstr("Print attachment");
      if (can_print(idx[menu->current]->content)) {
        STATE s;
        int status;

        endwin();
        mutt_create_filter(Print, &s.fpout, 0);
        s.fpin = Context->folder->fp;
        s.displaying = 0;
        s.prefix = 0;
        mutt_body_handler(idx[menu->current]->content, &s);
        fclose(s.fpout);
        wait(&status);
      }
      else
        mutt_error("I don't know how to print that!");
      break;

    case OP_ATTACH_SAVE_ENTRY:

      buffer[0] = 0;
      ci_get_field("Save to file: ", buffer, 0, CI_complete);
      if (buffer[0]) {
        if (mutt_save_attachment(idx[menu->current]->content, buffer) == 0)
          mutt_error("Attachment saved.");
        else
          mutt_error("Error while saving attachment!");
      }
      break;

    case OP_GENERIC_EXIT:

      while (count-- > 0) safe_free((void **)&idx[count]);
      safe_free((void **)&idx);
      mutt_menuDestroy(&menu);
      return;
    }
  }
  /* not reached */
}
