/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include "mutt.h"
#include "mutt_curses.h"
#include "rfc822.h"
#include "rfc1522.h"
#include "keymap.h"

#define DISPOSITION(X) X==DISPBODY?"attachment":"inline"

extern char *tspecials;

#ifdef ENABLE_RANDOM_SIG
extern char *addRandomSig (char *);
#endif /* ENABLE_RANDOM_SIG */

static int destination_is_local (ADDRESS *addr)
{
  while (addr) {
    if (strcasecmp(addr->host, Hostname) != 0 && strcasecmp(addr->host, Fqdn) != 0)
      return(0);
    addr = addr->next;
  }
  return(1);
}

static void append_signature (ENVELOPE *env, FILE *f)
{
  char buffer[LONG_STRING], *c;
  FILE *tmpfp;
  int len;

  if (Local_sig[0] || Remote_sig[0]) {
    if (destination_is_local(env->to) &&
        destination_is_local(env->cc) &&
        destination_is_local(env->bcc))
      c = Local_sig;
    else
      c = Remote_sig;
  }
  else
    c = Signature;

#ifdef ENABLE_RANDOM_SIG
  if ((c = addRandomSig(c)) == NULL) c = buffer;
#endif /* ENABLE_RANDOM_SIG */

  tmpfp = fopen(c, "r");
  if (tmpfp) {
    if (option(OPTSIGDASHES))
      fputs("\n-- \n", f); /* trailing space is recommended! */
    while ((len = fread(buffer, 1, sizeof(buffer), tmpfp)) > 0)
      fwrite(buffer, 1, len, f);
    fclose(tmpfp);
#ifdef ENABLE_RANDOM_SIG
  unlink(c);
#endif /* ENABLE_RANDOM_SIG */
  }
}

static BODY *mkfwdattach (char *folder, HEADER *h)
{
  BODY *tmpatt;
  char buffer[STRING];

  tmpatt = mutt_new_body();
  tmpatt->type = TYPEMESSAGE;
  tmpatt->subtype = safe_strdup("rfc822");
  switch(h->content->encoding) {
  case ENCBINARY:
    tmpatt->encoding = ENCBINARY;
    break;
  case ENC8BIT:
    tmpatt->encoding = ENC8BIT;
    break;
  default:
    tmpatt->encoding = ENC7BIT;
    break;
  }
  sprintf(buffer, "%s:%d", folder, h->msgno-1); /* msgno is cur+1 */
  tmpatt->filename = safe_strdup(buffer);
  strcpy(buffer, "Forwarded message from ");
  rfc822_write_address(buffer, h->from);
  tmpatt->description = safe_strdup(buffer);
  tmpatt->use_disp = 0;
  return tmpatt;
}

static int include_message (HEADER *cur, FILE *out)
{
  STATE *s;
  char buffer[STRING];
  
  s = mutt_new_state();
  s->fpin = Context->folder->fp;
  s->fpout = out;

  mutt_make_string(buffer, Attribution, cur, sizeof(buffer)-1);
  state_puts(buffer, s);
  state_putc('\n', s);

  mutt_make_string(buffer, Prefix, cur, sizeof(buffer)-1);
  s->prefix = buffer;

  if (option(OPTHEADER)) {
    fseek(s->fpin, cur->offset, 0);
    copy_header(s, 0);
  }

  mutt_body_handler(cur->content, s);
  if (PostIndentString[0]) {
    mutt_make_string(buffer, PostIndentString, cur, sizeof(buffer)-1);
    state_puts(buffer, s);
    state_putc('\n', s);
  }

  mutt_free_state(&s);
  return 0;
}

static ADDRESS *remove_user (ADDRESS *a)
{
  ADDRESS *top = 0, *last = 0;

  while (a) {
    if (!mutt_addr_is_user(a)) {
      if (top) {
        last->next = a;
        last = last->next;
      }
      else
        last = top = a;
      a = a->next;
      last->next = 0;
    }
    else {
      ADDRESS *tmp = a;
      
      a = a->next;
      tmp->next = 0;
      mutt_free_address(&tmp);
    }
  }
  return top;
}

static ADDRESS *find_mailing_lists (ADDRESS *t, ADDRESS *c)
{
  ADDRESS *top = 0, *ptr = t, *tmp;

  /*
   * Form the two lists into one for simpler processing.
   */
  if (t) {
    while (ptr->next) ptr = ptr->next;
    ptr->next = c;
  }
  else
    t = c;

  while (t) {
    if (mutt_is_mail_list(t)) {
      /*
       * rfc822_cpy_adr() copies the whole list, not just one element, so
       * temporarily cut the list.
       */
      tmp = t->next;
      t->next = 0;
      if (top) {
        ptr->next = rfc822_cpy_adr(t);
      }
      else
        ptr = top = rfc822_cpy_adr(t);
      t->next = tmp;
      while (ptr->next) ptr = ptr->next;
    }
    t = t->next;
  }
  return top;
}

static void append_list (ADDRESS **d, ADDRESS *s)
{
  ADDRESS *t;

  if (*d) {
    t = *d;
    while (t->next) t = t->next;
    t->next = rfc822_cpy_adr (s);
  }
  else
    *d = rfc822_cpy_adr (s);
}

static int edit_envelope (ENVELOPE *en)
{
  char buffer[STRING];
  int i;
  ADDRESS *tadr;

  buffer[0] = 0;
  if (en->to) rfc822_write_address(buffer, en->to);
  mutt_free_address(&en->to);
      
  if (buffer[0] == 0) {
    mvaddstr(LINES-2, 0, "To (? for aliases): ");
    refresh();
    if ((i = getch()) == '?') {
      if ((tadr = alias_menu()))  en->to = rfc822_cpy_adr(tadr);
      clear();
    }
    else if (CI_is_return(i) || i == ctrl('G'))
      return -1;
    else {
      buffer[0] = i;
      buffer[1] = 0;
    }
  }
      
  /*
   * If the user didn't select an alias, prompt for the address now.
   */
  if (!en->to) {
    ci_get_field("To: ", buffer, 0, 0);
    if (buffer[0] == 0) return(-1);
    rfc822_parse_adrlist(&en->to, buffer, "@" /* Fqdn */);
    en->to = mutt_expand_aliases(en->to);
  }
      
  buffer[0] = 0;
  rfc822_write_address(buffer, en->to);
  move(LINES-3, COLS/2);
  printw("To: %-*.*s", COLS/2-4, COLS/2-4, buffer);
  
  if (option(OPTASKCC)) {
    buffer[0] = 0;
    if (en->cc) rfc822_write_address(buffer, en->cc);
    mutt_free_address(&en->cc);
    ci_get_field("Cc: ", buffer, 0, 0);
    rfc822_parse_adrlist(&en->cc, buffer, "@" /* Fqdn */);
    en->cc = mutt_expand_aliases(en->cc);
  }
      
  if (en->subject)
    strcpy(buffer, en->subject);
  else
    buffer[0] = 0;
  ci_get_field("Subject: ", buffer, 0, 0);
  if (en->subject) safe_free((void **)&en->subject);
  if (!buffer[0] && !mutt_yesorno("No subject, continue?", 0)) return(-1);
  en->subject = safe_strdup(buffer);
  return 0;
}

void mutt_include_forward (HEADER *h, FILE *f)
{
  char buffer[LONG_STRING];

  buffer[0] = 0;
  rfc822_write_address(buffer, h->from);
  fprintf(f, "-----Forwarded message from %s-----\n\n", buffer);
  fseek(Context->folder->fp, h->offset, 0);
  while (fgets(buffer, sizeof(buffer)-1, Context->folder->fp) != NULL) {
    if (strncmp("From ", buffer, 5) != 0 &&
	strncmp("Status:", buffer, 7) != 0 &&
	strncmp("X-Status:", buffer, 9) != 0 &&
	strncasecmp("Content-Length:", buffer, 15) != 0)
      fputs(buffer, f);
    if (buffer[0] == '\n')
      break;
  }
  mutt_copy_bytes(MAILFP, f,
		  h->length - (ftell(Context->folder->fp) - h->offset));
  fputs("\n-----End of forwarded message-----\n", f);
}

#define BATCHMODE (flags == SENDBATCH)

static void process_user_defined_headers (METAENV *en)
{
  USERHEADER *uh = User_defined_headers;
  char buf[STRING];
  int i=0;
  ENVELOPE *env = en->env;

  while (uh) {
    if (strcasecmp("to", uh->key) == 0)
      rfc822_parse_adrlist(&env->to, uh->val, "@" /* Fqdn */);
    else if (strcasecmp("cc", uh->key) == 0)
      rfc822_parse_adrlist(&env->cc, uh->val, "@" /* Fqdn */);
    else if (strcasecmp("bcc", uh->key) == 0)
      rfc822_parse_adrlist(&env->bcc, uh->val, "@" /* Fqdn */);
    else if (strcasecmp("from", uh->key) == 0) {
      /*
       * User has specified a default From: address.  Get rid of the
       * default address.
       */
      mutt_free_address(&env->from);
      strcpy(buf, uh->val);
      rfc822_parse_adrlist(&env->from, buf, Fqdn);
    }
    else {
      sprintf(buf, "%s: %s", uh->key, uh->val);
      en->userhdrs[i++] = safe_strdup(buf);
    }
    uh = uh->next;
  }
  en->userhdrs[i] = 0;
}

LIST *mutt_copy_list (LIST *p)
{
  LIST *t, *r=NULL, *l=NULL;

  while (p) {
    t = (LIST *)safe_malloc(sizeof(LIST));
    t->data = safe_strdup(p->data);
    t->next = NULL;
    if (l) {
      r->next = t;
      r = r->next;
    }
    else
      l = r = t;
    p = p->next;
  }
  return(l);
}

static int envelope_defaults (int flags, HEADER *cur, ENVELOPE *en,
			      FILE *tempfp)
{
  ADDRESS *tmpaddr;
  char buffer[LONG_STRING];
  int i;

  if (flags & SENDREPLY) {
    if (flags & SENDLISTREPLY) {
      if ((en->to = find_mailing_lists(cur->to, cur->cc)) == 0) {
	mutt_error("Could not find any mailing lists in the message!");
	return(-1);
      }
    }
    else {
      if (cur->reply_to) {
	if (option(OPTASKREPLYTO)) {
	  /*
	   * There are quite a few mailing lists which set the Reply-To:
	   * header field to the list address, which makes it quite impossible
	   * to send a message to only the sender of the message.  This
	   * provides a way to do that.
	   */
	  strcpy(buffer, "Reply to ");
	  rfc822_write_address(buffer, cur->reply_to);
	  strcat(buffer, "? ");
	  if (mutt_yesorno(buffer, 1))
	    en->to = rfc822_cpy_adr(cur->reply_to);
	  else
	    en->to = rfc822_cpy_adr(cur->from);
	}
	else
	  en->to = rfc822_cpy_adr(cur->reply_to);
      }
      else
	en->to = rfc822_cpy_adr(cur->from);
    }

    if (cur->subject) {
      if (strncasecmp("re:", cur->subject, 3) == 0)
	en->subject = safe_strdup(cur->subject);
      else {
	en->subject = (char *)safe_malloc(strlen(cur->subject)+5);
	sprintf(en->subject, "Re: %s", cur->subject);
      }
    }

    if (flags & SENDGROUPREPLY) {
      en->cc = rfc822_cpy_adr(cur->to);
      if ((tmpaddr = en->cc)) {
	while (tmpaddr->next) tmpaddr = tmpaddr->next;
	tmpaddr->next = rfc822_cpy_adr(cur->cc);
      }
      else
	en->cc = rfc822_cpy_adr(cur->cc);
      if (! option(OPTMETOO)) en->cc = remove_user(en->cc);
    }

    mutt_make_string(buffer, In_reply_to, cur, sizeof(buffer)-1);
    en->in_reply_to = safe_strdup(buffer);

/* Ask copy message on reply - pms@romance.kaist.ac.kr */
strcpy(buffer, "Copy message? ");
if (mutt_yesorno(buffer, 1)) {
    if (Context->selected) {
      for (i = 0; i < Context->msgcount; i++)
	if (Context->hdrs[i]->selected) {
	  if (include_message(Context->hdrs[i], tempfp) == -1) {
	    mutt_error("Could not include all requested messages!");
	    return(-1);
	  }
	  fputc('\n', tempfp);
	}
    }
    else
      include_message(cur, tempfp);
}

    en->references = mutt_copy_list(cur->references);
    mutt_add_to_list(&en->references, cur->message_id);
  }
  else if (flags & SENDFORWARD) {
#if 0
    if (option(OPTMIMEFORWARD)) {
      if (Context->selected) {
	BODY *tmpatt = 0;

	for (i=0; i < Context->msgcount; i++) {
	  if (Context->hdrs[i]->selected) {
	    if (attach) {
	      tmpatt->next = mkfwdattach(Context->folder->name, Context->hdrs[i]);
	      tmpatt = tmpatt->next;
	    }
	    else
	      tmpatt = attach = mkfwdattach(Context->folder->name, Context->hdrs[i]);
	  }
	}
      }
      else
	attach = mkfwdattach(Context->folder->name, cur);
    }
    else
#endif
    {
      /*
       * Some people insist on using this old method of forwarding a message
       * so that they can edit it.
       */
      if (Context->selected) {
	for (i=0; i<Context->msgcount; i++) {
	  if (Context->hdrs[i]->selected)
	    mutt_include_forward(Context->hdrs[i], tempfp);
	}
      }
      else
	mutt_include_forward(cur, tempfp);
    }

    /* set the default subject for the message. */
    if (!cur->subject)
      en->subject = 0;
    else if (strncasecmp("fwd:", cur->subject, 4) == 0)
      en->subject = safe_strdup(cur->subject);
    else {
      en->subject = (char *)safe_malloc(strlen(cur->subject)+6);
      sprintf(en->subject, "Fwd: %s", cur->subject);
    }
  }

  if (flags == 0 || flags & SENDFORWARD || !option(OPTEDITHDRS)) {
    if (edit_envelope(en) == -1) return(-1);
  }
  
  return(0);
}

void mutt_free_metaenv (METAENV **p)
{
  int i;

  mutt_free_envelope(&(*p)->env);
  for (i=0; (*p)->userhdrs[i]; i++) safe_free((void **)&(*p)->userhdrs[i]);
  safe_free((void **)p);
}

int ci_send_message (METAENV *env, char *inc, int flags)
{
  char buffer[LONG_STRING], tempfile[_POSIX_PATH_MAX];
  int i, redraw = 1, done = 0, postponed=0;
  HEADER *cur = 0;
  ADDRESS *tmpaddr;
  BODY *attach = 0, *new;
  FILE *tempfp;
#ifdef _PGPPATH
  int pgpaction = 0;
  char *pgpkeylist = 0;
#endif
  ENVELOPE *en;

  sprintf(tempfile, "%s/mutt.XXXXXX", Tempdir);
  mktemp(tempfile);

  if ((tempfp = fopen(tempfile, "w")) == NULL) {
    dprint(1,(debugfile, "newsend_message: can't create tempfile %s (errno=%d)\n", tempfile, errno));
    mutt_error("Can't open temporary file!");
    return -1;
  }

  /* If the user specified a file to include on the command line, do so now */
  if (inc && access(inc, R_OK) != -1) {
    FILE *f = fopen(inc, "r");

    if (f) {
      while ((i = fread(buffer, 1, sizeof(buffer), f)) > 0)
	fwrite(buffer, 1, i, tempfp);
      fclose(f);
    }
  }

  if (env) {
    env->env->to = mutt_expand_aliases(env->env->to);
    env->env->cc = mutt_expand_aliases(env->env->cc);
    env->env->from = mutt_default_from();
  }
  else {
    env = mutt_new_metaenv();

    if (flags == 0) {
      /*
       * If the user is composing a new message, check to see if there are any
       * postponed messages first.
       */
      flags = ci_get_postponed_message(tempfp, env, &cur);
      if (flags < 0)
	flags=0;
      else
	postponed=1;
    }
  }

  if (!postponed) {
    if (!env->env) {
      env->env = mutt_new_envelope();
      env->env->from = mutt_default_from();
    }

    if (!BATCHMODE) {

      /* Find the current message */
      if (Context && Context->hdrs) {
	if (Context->selected) {
	  for (i = 0; i < Context->msgcount; i++)
	    if (Context->hdrs[i]->selected) {
	      cur = Context->hdrs[i];
	      break;
	    }
	}
	else
	  cur = Context->hdrs[ *Context->v2r[Context->current] ];
      }

      if (envelope_defaults(flags, cur, env->env, tempfp) == -1) {
	fclose(tempfp);
	unlink(tempfile);
	mutt_free_metaenv(&env);
	return(-1);
      }
    }
    else { /* BATCHMODE */
      while ((i = fread(buffer, 1, sizeof(buffer), stdin)) > 0)
	fwrite(buffer, 1, i, tempfp);
    }

    process_user_defined_headers(env);
    append_signature(env->env, tempfp);
  }
  fclose(tempfp);

  if (!BATCHMODE) {
    if (option(OPTEDITHDRS))
      ci_edit_headers(tempfile, env);
    else
      ci_edit_file(tempfile);
  }
  else
    done = 1; /* don't go into the loop. */

  en = env->env;

#ifdef _PGPPATH
  main_loop:
#endif /* _PGPPATH */

  while (!done) {
    if (redraw) {
#define HDR_TO 2
#define HDR_CC 3
#define HDR_BCC 4
#define HDR_SUBJECT 5
#define HDR_USER 7
#define HDR_BODY_MSG 9
#define HDR_XOFFSET 9
      clear();
      CenterLine("Send Menu", 0);
      buffer[0] = 0;
      rfc822_write_address(buffer, env->env->to);
      move (HDR_TO, 0);
      printw("To:      %-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET,buffer);
      buffer[0] = 0;
      rfc822_write_address(buffer, en->cc);
      move (HDR_CC, 0);
      printw("Cc:      %-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET,buffer);
      buffer[0] = 0;
      rfc822_write_address(buffer, en->bcc);
      move (HDR_BCC, 0);
      printw ("Bcc:     %-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET,buffer);
      move (HDR_SUBJECT, 0);
      printw ("Subject: %-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET,en->subject?en->subject:"");
      if (attach) mvaddstr(HDR_BODY_MSG, 0, "This message has attachment(s).");
      redraw = 0;
    }

/* PMS */
    mvaddstr(LINES-3, 0, "Send message? ([y]/n/?): ");
    refresh();

    i = keymap_lookup(getch(), MENU_SEND);

    switch (i) {
    case OP_SEND_EDIT_TO:

      buffer[0] = 0;
      rfc822_write_address(buffer, en->to);
      ci_enter_string(buffer, HDR_TO, HDR_XOFFSET, 0, 0);
      mutt_free_address(&en->to);
      rfc822_parse_adrlist(&en->to, buffer, "@" /* Fqdn */);
      en->to = mutt_expand_aliases(en->to);
      buffer[0] = 0;
      rfc822_write_address(buffer, en->to);
      move(HDR_TO, HDR_XOFFSET);
      clrtoeol();
      printw("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buffer);
      break;

    case OP_SEND_EDIT_SUBJECT:

      if (en->subject)
        strcpy(buffer, en->subject);
      else
        buffer[0] = 0;
      ci_enter_string(buffer, HDR_SUBJECT, HDR_XOFFSET, 0, 0);
      if (en->subject) safe_free((void **)&en->subject);
      en->subject = safe_strdup(buffer);
      break;

    case OP_SEND_EDIT_BCC:

      buffer[0] = 0;
      rfc822_write_address(buffer, en->bcc);
      ci_enter_string(buffer, HDR_BCC, HDR_XOFFSET, 0, 0);
      mutt_free_address(&en->bcc);
      rfc822_parse_adrlist(&en->bcc, buffer, "@" /* Fqdn */);
      en->bcc = mutt_expand_aliases(en->bcc);
      buffer[0] = 0;
      rfc822_write_address(buffer, en->bcc);
      move(HDR_BCC, HDR_XOFFSET);
      clrtoeol();
      printw("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buffer);
      break;

    case OP_SEND_EDIT_CC:

      buffer[0] = 0;
      rfc822_write_address(buffer, en->cc);
      ci_enter_string(buffer, HDR_CC, HDR_XOFFSET, 0, 0);
      mutt_free_address(&en->cc);
      rfc822_parse_adrlist(&en->cc, buffer, "@" /* Fqdn */);
      en->cc = mutt_expand_aliases(en->cc);
      buffer[0] = 0;
      rfc822_write_address(buffer, en->cc);
      move(HDR_CC, HDR_XOFFSET);
      clrtoeol();
      printw("%-*.*s", COLS-HDR_XOFFSET, COLS-HDR_XOFFSET, buffer);
      break;


    case OP_SEND_EDIT_MESSAGE:

      if (option(OPTEDITHDRS)) {
        ci_edit_headers(tempfile, env);
        redraw = 1; /* headers may have changed! */
      }
      else
        ci_edit_file(tempfile);
      break;

    case OP_SEND_EDIT_HEADERS:

      ci_edit_headers(tempfile, env);
      redraw = 1;
      break;

    case OP_SEND_ATTACH_FILE:

      attach = ci_attach_menu(attach);
      redraw = 1;
      break;

    case OP_SEND_ALIAS_MENU:

      /*
       * Pop up a menu with the user's aliases in it.  If they selected one,
       * allow them to add it to any of the recipient lists.
       */
      if ((tmpaddr = alias_menu())) {
	i = 0;
	while (i == 0) {
	  mvaddstr (LINES-2, 0, "Add to (t)o, (c)c, (b)cc or (f)orget it? ");
	  refresh();
	  switch (i = getch()) {
	  case 't':
	    append_list(&en->to, tmpaddr);
	    break;
	  case 'c':
	    append_list(&en->cc, tmpaddr);
	    break;
	  case 'b':
	    append_list(&en->bcc, tmpaddr);
	    break;
	  case 'f':
	    break;
	  default:
	    beep ();
	    i = 0;
	    break;
	  }
	}
      }
      redraw = 1;
      break;

    case OP_SEND_SEND_MESSAGE:

      done = 1;
      break;

    case OP_SEND_ABORT_MESSAGE:

      if (!option(OPTASKPOSTPONE) || !mutt_yesorno("Postpone this message? ", FALSE)) {
	unlink(tempfile);
	mutt_free_body(&attach);
	mutt_free_metaenv(&env);
	mutt_error("Mail not sent.");
	return (-1);
      }

      /* fall through to postpone! */

    case OP_SEND_POSTPONE_MESSAGE:
      mutt_postpone_message(tempfile, env, flags & SENDREPLY ? cur->message_id : NULL);
      unlink(tempfile);
      mutt_free_metaenv(&env);
      mutt_free_body(&attach);
      mutt_error("Message postponed.");
      return(-1);
      break;

    case OP_SEND_REFRESH:

      redraw = 1;
      break;

    case OP_SEND_HELP_MENU:

      if (help(MENU_SEND)) redraw = 1;
      break;

#ifdef ISPELL
    case OP_SEND_ISPELL:

      endwin();
      sprintf(buffer, "%s %s", ISPELL, tempfile);
      mutt_system(buffer);
      break;
#endif /* ISPELL */

#if 0
#ifdef MIXMASTER_PATH
    case OP_SEND_MIXMASTER_MENU:

      remail = mix_menu();
      redraw = 1;
      break;
#endif /* MIXMASTER_PATH */
#endif
 
#ifdef _PGPPATH
    case OP_SEND_PGP_MENU:

      pgpaction = pgp_sendMenu();
      break;
#endif /* _PGPPATH */

    default:

      mutt_error("Key is not bound!");
      beep();
    }
  }

#ifdef _PGPPATH
  /* Do a quick check to make sure that we can find all of the encryption
   * keys if the user has requested this service.
   */
  if (pgpaction & PGPENCRYPT) {
    if ((pgpkeylist = pgp_findKeys (en->to, en->cc, en->bcc)) == 0)
      goto main_loop;
  }
#endif

  /* Create a record for the main body of this message. */
  new = mutt_make_attach(tempfile);

  /* now add any attachments. */
  if (attach)
    attach = mutt_make_multipart(new, attach);
  else
    attach = new;

#ifdef _PGPPATH
  if (pgpaction & PGPENCRYPT) {
    attach = pgp_encryptMessage(attach, pgpkeylist, pgpaction & PGPSIGN);
    safe_free((void **)&pgpkeylist);
  }
  else if (pgpaction == PGPSIGN)
    attach = pgp_signMessage(attach);
#endif

  mutt_send_message(env, attach, BATCHMODE);

  /* now free up the memory used to generate this message. */
  mutt_free_body(&attach);
  mutt_free_metaenv(&env);

  if (flags & SENDREPLY) {
    if (!postponed && Context->selected) {
      for (i = 0; i < Context->msgcount; i++) {
	if (Context->hdrs[i]->selected && !Context->hdrs[i]->replied) {
	  Context->hdrs[i]->replied = 1;
	  Context->hdrs[i]->changed = 1;
	}
      }
    }
    else if (cur && !cur->replied) {
      cur->replied = 1;
      cur->changed = 1;
    }
  }

  return 0;
}
