/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <stdlib.h>
#include "muttlib.h"
#include "rfc822.h"

#define USE_QSORT

static int days_per_month[12] = {
	31, 28, 31, 30, 31, 30, 30, 31, 30, 31, 30, 31
};

#define is_leap_year(x) ((x)%4 == 0 && ((x)%100 != 0))

/* return the number of secs since Jan 1, 1971 12:00am GMT */
static long mygmtime (HEADER *h) {
  long t = 0;
  int i;

  for (i=0; i<h->year; i++) {
    t += 365 * 24 * 3600;
    if (is_leap_year(i+1971)) t += 24 * 3600;
  }
  for (i=0; i<h->month; i++) {
    t += days_per_month[i] * 24 * 3600;
  }
  t += h->day * 24 * 3600;
  t += h->hours * 3600;
  t += h->minutes * 60;
  t += h->seconds;
  if (h->zoccident) { /* west of GMT */
    t += h->zhours * 3600;
    t += h->zminutes * 60;
  }
  else { /* east of GMT */
    t -= h->zhours * 3600;
    t -= h->zminutes * 60;
  }
  return t;
}

#ifdef USE_QSORT

#define SORTCODE(x) Sort & SORTREVERSE ? -(x) : x

static int compare_size (const void *a, const void *b)
{
  HEADER **ppa = (HEADER **)a;
  HEADER **ppb = (HEADER **)b;
  HEADER *pa = *ppa, *pb = *ppb;
  int rc;

  if (pa->length < pb->length)
    rc = -1;
  else if (pa->length == pb->length)
    rc = 0;
  else
    rc = 1;
  return(SORTCODE(rc));
}

static int compare_date_sent (const void *a, const void *b)
{
  HEADER **ppa = (HEADER **)a;
  HEADER **ppb = (HEADER **)b;
  HEADER *pa = *ppa, *pb = *ppb;
  time_t at, bt;
  int rc;
  
  at = mygmtime(pa);
  bt = mygmtime(pb);
  if (at < bt)
    rc = -1;
  else if (at == bt)
    rc = 0;
  else
    rc = 1;
  return (SORTCODE(rc));
}

static int compare_subject (const void *a, const void *b)
{
  HEADER **ppa = (HEADER **)a;
  HEADER **ppb = (HEADER **)b;
  HEADER *pa = *ppa, *pb = *ppb;
  int rc;

  if (!pa->subject) {
    if (!pb->subject)
      rc = compare_date_sent(pa, pb);
    else
      rc = -1;
  }
  else if (!pb->subject)
    rc = 1;
  else
   rc = strcmp(pa->subject, pb->subject);
  return(SORTCODE(rc));
}

static int compare_from (const void *a, const void *b)
{
  HEADER **ppa = (HEADER **)a;
  HEADER **ppb = (HEADER **)b;
  HEADER *pa = *ppa, *pb = *ppb;
  char buf1[STRING], buf2[STRING];

  buf1[0] = 0;
  buf2[0] = 0;
  rfc822_write_address(buf1, pa->from);
  rfc822_write_address(buf2, pb->from);
  return(SORTCODE(strcmp(buf1,buf2)));
}

static int compare_date_received (const void *a, const void *b)
{
  HEADER **ppa = (HEADER **)a;
  HEADER **ppb = (HEADER **)b;
  HEADER *pa = *ppa, *pb = *ppb;
  int rc;
  
  if (pa->received < pb->received)
    rc = -1;
  else if (pa->received == pb->received) /* this can never happen! */
    rc = 0;
  else
    rc = 1;
  return(SORTCODE(rc));
}

#else /* !USE_QSORT */

/* Returns TRUE if "a" comes before "b" */
static int comes_before (HEADER *a, HEADER *b, int magic) {
  char buf1[STRING], buf2[STRING];
  short rc;

  switch (magic) {
  case SORTSIZE:
    if (a->length < b->length) return 1;
    break;
  case SORTDATE:
    if (mygmtime(a) < mygmtime(b)) return 1;
    break;
  case SORTSUBJECT:
    if (!a->subject) return 1;
    if (!b->subject) return 0;
    if ((rc = strcasecmp(a->subject, b->subject)) < 0) return 1;
    else if (rc == 0) return(comes_before(a, b, SORTDATE));
    break;
  case SORTFROM:
    rfc822_write_address(buf1, a->from);
    rfc822_write_address(buf2, b->from);
    if ((rc = strcasecmp(buf1, buf2)) < 0) return 1;
    else if (rc == 0) return(comes_before(a, b, SORTDATE));
    break;
  case SORTORDER:
    if (a->received < b->received) return 1;
    break;
  }
  return 0;
}

#define COMPARE Sort & SORTREVERSE ? \
                 comes_before(newtop[j], ctx->hdrs[i], Sort & SORTMASK) : \
                 comes_before(ctx->hdrs[i], newtop[j], Sort & SORTMASK)

#endif /* USE_QSORT */

void mutt_sort_headers (CONTEXT *ctx)
{
  int i, current;

#ifdef USE_QSORT
  int (*f)(const void *, const void *);
  
  switch(Sort & SORTMASK) {
  case SORTDATE:
    f = compare_date_sent;
    break;
  case SORTFROM:
    f = compare_from;
    break;
  case SORTSUBJECT:
    f = compare_subject;
    break;
  case SORTORDER:
    f = compare_date_received;
    break;
  case SORTSIZE:
    f = compare_size;
    break;
  default:
    mutt_error("Error in mutt_sort_headers()...  This should not happen!");
    return;
  }
  qsort((void *)ctx->hdrs, ctx->msgcount, sizeof(HEADER *), f);
#else
  HEADER **newtop;
  int newtopmax = 0;
  int j, k;
#ifdef DEBUG
  int true=0, compare=0;
#endif

  if (ctx->msgcount == 0) return;

  newtop = (HEADER **)safe_malloc(sizeof(HEADER *) * ctx->hdrmax);
  newtop[0] = ctx->hdrs[0];
  newtopmax++;

  for (i=1; i < ctx->msgcount; i++) {
    for (j=0; j < newtopmax; j++) {
      if (COMPARE) {
	for (k=newtopmax; k>j; k--)
	  newtop[k] = newtop[k-1];
	newtop[j] = ctx->hdrs[i];
	newtopmax++;
#ifdef DEBUG
	true++; /* keep track of the number of true comparisons. */
#endif
	break;
      }
#ifdef DEBUG
      compare++; /* keep track of the number of comparisons */
#endif
    }
    if (j == newtopmax)
      newtop[newtopmax++] = ctx->hdrs[i];
  }
  dprint(1,(debugfile, "mutt_sort_headers(): # compare=%d, # true=%d.\n",
	    compare, true));
  safe_free((void **)&ctx->hdrs);
  ctx->hdrs = newtop;
#endif /* USE_QSORT */

  current = ctx->current;
  ctx->current = -1;
  for (i = 0; i < ctx->msgcount; i++) {
    /*
     * what used to be the current message no. has probably changed, so we
     * need to update it here.
     */
    if (ctx->current == -1 && ctx->hdrs[i]->msgno - 1 == current)
      ctx->current = i;

    ctx->hdrs[i]->virtual = i;
    ctx->hdrs[i]->msgno = i + 1;
    *ctx->v2r[i] = i; /* reset the virtual message mapping. */
  }
  ctx->vcount = ctx->msgcount;
}
